"""
    Module with dispute related entities
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T, 
    Money, 
    ActionLink, 
    PayPalEntity, 
    ResponseType, 
    PaypalMessage, 
    PaypalMerchant, 
    PatchUpdateRequest 
)

class DisputeState(Enum):
    """Enumeration of dispute states
    """
    # Filters the disputes in the response to those with the REQUIRED_ACTION dispute state.
    REQUIRED_ACTION =  1
    # Filters the disputes in the response to those with the REQUIRED_OTHER_PARTY_ACTION dispute state.
    REQUIRED_OTHER_PARTY_ACTION =  2
    # Filters the disputes in the response to those with the UNDER_PAYPAL_REVIEW dispute state.
    UNDER_PAYPAL_REVIEW = 3 
    # Filters the disputes in the response to those with the RESOLVED dispute state.
    RESOLVED = 4
    # Filters the disputes in the response to those with the OPEN_INQUIRIES dispute state.
    OPEN_INQUIRIES = 5
    # Filters the disputes in the response to those with the APPEALABLE dispute state.
    APPEALABLE = 6

class DisputeStatus(Enum):
    """Enumeration for the possible dispute status
    """
    # The dispute is open.
    OPEN = 1
    # The dispute is waiting for a response from the customer.
    WAITING_FOR_BUYER_RESPONSE = 2
    # The dispute is waiting for a response from the merchant.
    WAITING_FOR_SELLER_RESPONSE = 3
    # The dispute is under review with PayPal.
    UNDER_REVIEW = 4
    # The dispute is resolved.
    RESOLVED = 5
    # The default status if the dispute does not have one of the other statuses.
    OTHER = 6

class DisputeReason(Enum):
    """Enumeration with known dispute reasons
    """
    # The customer did not receive the merchandise or service.
    MERCHANDISE_OR_SERVICE_NOT_RECEIVED = 1
    # The customer reports that the merchandise or service is not as described.
    MERCHANDISE_OR_SERVICE_NOT_AS_DESCRIBED = 2
    # The customer did not authorize purchase of the merchandise or service.
    UNAUTHORISED = 3
    # The refund or credit was not processed for the customer.
    CREDIT_NOT_PROCESSED = 4
    # The transaction was a duplicate.
    DUPLICATE_TRANSACTION = 5
    # The customer was charged an incorrect amount.
    INCORRECT_AMOUNT = 6
    # The customer paid for the transaction through other means.
    PAYMENT_BY_OTHER_MEANS = 7
    # The customer was being charged for a subscription or a recurring transaction that was canceled.
    CANCELED_RECURRING_BILLING = 8
    # A problem occurred with the remittance.
    PROBLEM_WITH_REMITTANCE = 9
    # Other.
    OTHER = 10

class DisputeChannel(Enum):
    """Enumeration with known dispute channels
    """
    # The customer contacts PayPal to file a dispute with the merchant.
    INTERNAL = 1
    # The customer contacts their card issuer or bank to request a refund.
    EXTERNAL = 2

class DisputeLifeCycleStage(Enum):
    """Enumeration with known dispute life cycle stages
    """
    # A customer and merchant interact in an attempt to resolve a dispute without escalation to PayPal. Occurs when the customer:
    #     Has not received goods or a service.
    #     Reports that the received goods or service are not as described.
    #     Needs more details, such as a copy of the transaction or a receipt.
    INQUIRY = 1
    # A customer or merchant escalates an inquiry to a claim, which authorizes PayPal to investigate the case and make a determination. Occurs only when the dispute channel is INTERNAL. This stage is a PayPal dispute lifecycle stage and not a credit card or debit card chargeback. All notes that the customer sends in this stage are visible to PayPal agents only. The customer must wait for PayPal’s response before the customer can take further action. In this stage, PayPal shares dispute details with the merchant, who can complete one of these actions:
    #         Accept the claim.
    #         Submit evidence to challenge the claim.
    #         Make an offer to the customer to resolve the claim.
    CHARGEBACK = 2
    # The first appeal stage for merchants. A merchant can appeal a chargeback if PayPal's decision is not in the merchant's favor.
    # If the merchant does not appeal within the appeal period, PayPal considers the case resolved.
    PRE_ARBITRATION = 3
    # The second appeal stage for merchants. A merchant can appeal a dispute for a second time if the first appeal was denied. 
    # If the merchant does not appeal within the appeal period, the case returns to a resolved status in pre-arbitration stage.
    ARBITRATION = 4

class Buyer(PayPalEntity):
    """Dispute buyer obj representation
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class ItemInfo(PayPalEntity):
    """Dispute item info obj representation
    """

    _ENTITY_TYPES = { 'dispute_amount': Money }

    def __init__(
            self, item_id: str = None, item_description: str = None, 
            partner_transaction_id: str = None, reason: str = None, 
            dispute_amount: Money = None, notes: str = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.notes = notes
        self.reason = reason
        self.item_id = item_id
        self.dispute_amount = dispute_amount
        self.item_description  = item_description
        self.partner_transaction_id = partner_transaction_id

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response = json_data, response_type = response_type)

class DisputeTransaction(PayPalEntity):
    """Transaction object representation
    """
    
    _ARRAY_TYPES = { 'messages': PaypalMessage, 'items': ItemInfo }
    _ENTITY_TYPES = { 'buyer': Buyer, 'gross_amount': Money, 'seller': PaypalMerchant }

    def __init__(
            self, buyer_transaction_id: str = None, seller_transaction_id: str = None, 
            transaction_status: str = None, buyer: Buyer = None, gross_amount: Money = None,
            seller: PaypalMerchant = None, messages: List[PaypalMessage] = [],
            invoice_number: str = None, custom: str = None, items: List[ItemInfo] = [], **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.buyer  = buyer
        self.seller = seller
        self.custom = custom
        self.items = items or []
        self.messages = messages or []
        self.gross_amount = gross_amount
        self.invoice_number = invoice_number
        self.transaction_status = transaction_status
        self.buyer_transaction_id = buyer_transaction_id
        self.seller_transaction_id = seller_transaction_id
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
    
    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES, cls._ARRAY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)
 
class DisputeUpdateRequest(PatchUpdateRequest):
    """Update request for PATCH dispute updates
    """
    
    def __init__(self, merchant_email:str, note: str, operation: str='add', path: str='/communication_details'):
        super().__init__(path, {'email': merchant_email, 'note': note}, operation)

class DisputeTracker(PayPalEntity):
    """Dispute tracking info
    """
    def __init__(self, tracking_number: str, carrier_name: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.carrier_name = carrier_name
        self.tracking_number = tracking_number
        self.tracking_url = kwargs.get('tracking_url')
        self.carrier_name_other = kwargs.get('carrier_name_other')

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['tracking_number'], json_data['carrier_name'],  
            json_response= json_data, response_type = response_type
        )

class RefundId(PayPalEntity):
    """Refund id obj representation
    """
    def __init__(self, refund_id: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.refund_id = refund_id
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('refund_id'), json_response = json_data, response_type = response_type)

class DisputeEvidenceInfo(PayPalEntity):
    """Dispute evidence info
    """

    _ARRAY_TYPES = { 'trackers': DisputeTracker, 'refund_ids': RefundId }

    def __init__(self, trackers: List[DisputeTracker] = [], refund_ids: List[RefundId]=[], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.tracking_info = trackers or []
        self.refunds_ids = refund_ids or []

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, dict(), cls._ARRAY_TYPES)    
        return cls(**args, json_response= json_data, response_type = response_type)

class Document(PayPalEntity):
    """Dispute document obj representation
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class DisputeEvidence(PayPalEntity):
    """Dispute evidence for requests
    """

    _ENTITY_TYPES = { 'document': Document, 'evidence_info': DisputeEvidenceInfo }

    def __init__(
        self, evidence_type: str = None, notes: str = None, document: Document = None,
        item_id: str = None, evidence_info: DisputeEvidenceInfo = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.notes = notes
        self.item_id = item_id
        self.document = document
        self.evidence_type = evidence_type
        self.evidence_info = evidence_info

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = super()._build_args(json_data, cls._ENTITY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)

class DisputeOutcome(PayPalEntity):
    """Dispute outcome object representation    
    """
    def __init__(self, outcome_code: str, amount_refunded: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.outcome_code = outcome_code
        self.amount_refunded = amount_refunded

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount = Money.serialize_from_json(json_data['amount_refunded'])
        
        return cls(
            json_data['outcome_code'], amount, json_response= json_data, response_type = response_type
        )

class Dispute(PayPalEntity):
    """Dispute object representation
    """

    _ARRAY_TYPES = { 'messages': PaypalMessage, 'disputed_transactions': DisputeTransaction }
    _ENTITY_TYPES = { 'offer': Money, 'dispute_amount': Money, 'dispute_outcome': DisputeOutcome }

    def __init__(
        self, dispute_id: str = None, reason: str = None, status: str = None, amount: Money = None, 
        offer: Money = None, dispute_outcome: DisputeOutcome = None, messages: List[PaypalMessage] = [], 
        disputed_transactions: List[DisputeTransaction] = [], dispute_channel: str = None, 
        dispute_state: str = None, dispute_life_cycle_stage: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.offer = offer
        self.reason = reason
        self.status = status
        self.dispute_id = dispute_id
        self.dispute_amount = amount
        self.messages = messages or []
        self.dispute_state = dispute_state
        self.dispute_channel = dispute_channel
        self.dispute_outcome = dispute_outcome
        self.disputed_transactions = disputed_transactions or []
        self.dispute_life_cycle_stage = dispute_life_cycle_stage
        self.extensions = self._json_response.get('extensions', kwargs.get('extensions')) 
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time')) 
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time')) 
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def dispute_state_enum(self) -> DisputeState:
        try:
            return DisputeState[self.dispute_state] if self.dispute_state else None
        except:
            return None
    
    @property
    def reason_enum(self) -> DisputeReason:
        try:
            return DisputeReason[self.reason] if self.reason else None
        except:
            return None

    @property
    def status_enum(self) -> DisputeStatus:
        try:
            return DisputeStatus[self.status] if self.status else None
        except:
            return None

    @property
    def dispute_channel_enum(self) -> DisputeChannel:
        try:
            return DisputeChannel[self.dispute_channel] if self.dispute_channel else None
        except:
            return None       
    
    @property
    def dispute_life_cycle_stage_enum(self) -> DisputeLifeCycleStage:
        try:
            return DisputeLifeCycleStage[self.dispute_life_cycle_stage] if self.dispute_life_cycle_stage else None
        except:
            return None

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        args = super()._build_args(json_data, cls._ENTITY_TYPES, cls._ARRAY_TYPES)
        return cls(**args, json_response= json_data, response_type = response_type)


"""
    Invoice api client for paypal REST resources.

    Resource docs & Reference: https://developer.paypal.com/docs/api/invoicing/v2/
"""

import json
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase
from pypaypal.errors import PaypalRequestError

from pypaypal.entities.invoicing.template import Template, InvoiceListRequestField
from pypaypal.entities.invoicing.invoice import Invoice, PaymentDetail, InvoiceSearchRequest

from pypaypal.entities.base import ( 
    PaypalPage,
    ActionLink,
    RefundDetail,
    ResponseType,
    PaypalApiResponse,
    PayPalErrorDetail,
    PaypalApiBulkResponse
)

from pypaypal.http import ( 
    parse_url,
    PayPalSession,     
    LIVE_API_BASE_URL,
    SANDBOX_API_BASE_URL     
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LIVE_API_BASE_URL, 'invoicing')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(SANDBOX_API_BASE_URL, 'invoicing')

T = TypeVar('T', bound = 'InvoiceClient')

I = TypeVar('T', bound = 'InvoiceTemplateClient')

class InvoiceClient(ClientBase):
    """Invoice v2 API client class
    """

    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)
    
    def generate_invoice_number(self) -> str:
        """Calls the paypal API to generate the next invoice number that is available to the merchant.
        
        Returns:
            str -- The generated invoice number

        Raises:
            PaypalRequestError -- If there's an error with the API request
        """
        api_response = self._session.post(parse_url(self._base_url, 'generate-next-invoice-number'), None)

        if api_response.status_code != 200:
            raise PaypalRequestError(PayPalErrorDetail.serialize_from_json(api_response.json()))
        
        return api_response.json()['invoice_number']

    def create_draft_invoice(self, invoice: Invoice) -> PaypalApiResponse:
        """Calls the paypal API to generate an invoice draft.
        
        Arguments:            
            invoice {Invoice} -- the invoice information

        Returns:
            PaypalApiResponse -- The paypal API response
        """
        response = self._session.post(parse_url(self._base_url, 'invoices'), json.dumps(invoice.to_dict()))

        if response.status_code != 201:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def list_invoices(self, page: int, page_size: int, total_required: bool, fields: List[str]) -> PaypalPage[Invoice]:
        """Calls the paypal API to get an invoice page.
        
        Arguments:
            page {int} -- current page
            page_size {int} -- page size 
            total_required {bool} -- total count required
            fields {List[str]} -- fields to be searched

        Returns:
            PaypalPage -- The paged elements in paypal API paged response 
        """
        query_params = { page: page, page_size: page_size, total_required: total_required, fields: ','.join(fields) }

        response = self._session.get(parse_url(self._base_url, 'invoices'), query_params)

        if response.status_code != 200:
            return PaypalPage.error(response)
        
        json_response = response.json()
        items = [ Invoice.serialize_from_json(x) for x in json_response['items']]
        links = [ ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in json_response['links'] ]

        return PaypalPage.success(response, json_response.get('total_items'), json_response.get('total_pages'), items, links)

    def delete_invoice(self, invoice_id: str) -> PaypalApiResponse:
        """Calls the paypal API to delete a draft or scheduled invoice, by ID.
           only invoices in the draft or scheduled state can be deleted. 
           For invoices that have already been sent, you can cancel the invoice. 
           After you delete a draft or scheduled invoice, you can no longer use it 
           or show its details. However, its possible to reuse the invoice number.
        
        Arguments:
            invoice_id {str} -- invoice id

        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """

        response = self._session.delete(parse_url(self._base_url, 'invoices', invoice_id))

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def fully_update_invoice(self, invoice_id: str, invoice: Invoice, send_to_recipient: bool = True, send_to_invoicer: bool = True) -> PaypalApiResponse[Invoice]:
        """Calls the paypal API to fully update an invoice. This call does not support partial updates
        
        Arguments:
            invoice_id {str} -- invoice id
            invoice {Invoice} -- invoice with the updated info
        
        Keyword Arguments:
            send_to_recipient {bool} --  Indicates whether to send the invoice update notification to the recipient. (default: {True})
            send_to_invoicer {bool} --   Indicates whether to send the invoice update notification to the merchant. (default: {True})

        Returns:
            PaypalApiResponse -- Api operation response status with parsed objects within
        """
        params = { 'send_to_recipient': send_to_recipient, 'send_to_invoicer': send_to_invoicer }

        response = self._session.put(
            parse_url(self._base_url, 'invoices', invoice_id),
            json.dumps(invoice.to_dict()),
            params = params
        )

        # There seems to be a copy/paste mistake in the docs
        # It says no content but has a response within it.
        if response.status_code // 100 != 2:
            return PaypalApiResponse.error(response)

        json_response = response.json()

        if json_response:
            return PaypalApiResponse.success(response, Invoice.serialize_from_json(json_response))

        return PaypalApiResponse.success(response)


    def show_invoice_details(self, invoice_id: str) -> PaypalApiResponse[Invoice]:
        """Calls the paypal API to show details for an invoice.
        
        Arguments:
            invoice_id {str} -- invoice id

        Returns:
            PaypalApiResponse -- Api operation response status with parsed objects within
        """
        response = self._session.get(parse_url(self._base_url, 'invoices', invoice_id))

        if response.status_code != 200:
            return PaypalApiResponse.error(response)

        return PaypalApiResponse.success(response, Invoice.serialize_from_json(response.json()))

    def cancel_sent_invoice(
        self, invoice_id: str, subject: str, note: str, send_to_invoicer: bool,
        send_to_recipient: bool, additional_recipients: List[str] = []) -> PaypalApiResponse[Invoice]:
        """Calls the paypal API to cancel a sent invoice, and, optionally, send a notification about the cancellation 
           to the payer, merchant, and CC: emails.
        
        Arguments:
            invoice_id {str} -- invoice id
            subject {str} --  The subject of the email that is sent as a notification to the recipient.
            note {str} -- A note to the payer.
            send_to_invoicer {bool} -- Indicates whether to send a copy of the email to the merchant.
            send_to_recipient {bool} -- Indicates whether to send a copy of the email to the recipient.
            additional_recipients {List[str]} -- An array of one or more CC: emails to which notifications are sent.
             If ignored, a notification is sent to all CC: email addresses that are part of the invoice.

        Returns:
            PaypalApiResponse -- Api operation response status with parsed objects within
        """
        url = parse_url(self._base_url, 'invoices', invoice_id, 'cancel')

        body = json.dumps({
            'subject': subject,
            'note': note,
            'send_to_invoicer': send_to_invoicer,
            'send_to_recipient': send_to_recipient
        })

        if additional_recipients:
            body['additional_recipients'] = [{'email_address': x} for x in additional_recipients]

        response = self._session.post(url, body)

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def generate_qr_code(self, invoice_id: str, width: int = 500, height: int = 500, action: str = 'pay') -> PaypalApiResponse:
        """ Calls the paypal API to generate a QR code for an invoice. 
            The QR code is a PNG image in Base64-encoded format that corresponds to the invoice ID. 
            You can generate a QR code for an invoice and add it to a paper or PDF invoice. 
            When customers use their mobile devices to scan the QR code, they are redirected to the PayPal mobile payment flow where they can view the invoice and pay online with PayPal or a credit card.
            Before you get a QR code, you must create an invoice and send an invoice to move the invoice from a draft to payable state. 
            Do not include an email address if you do not want the invoice emailed.
        
        Arguments:
            invoice_id {str} -- existing invoice id
        
        Keyword Arguments:
            width {int} -- The width, in pixels, of the QR code image. Value is from 150 to 500 (default: {500})
            height {int} -- The height, in pixels, of the QR code image. Value is from 150 to 500 (default: {500})
            action {str} -- The type of URL for which to generate a QR code. Valid values are 'pay' and 'details'. (default: {'pay'})
        
        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """
        url = parse_url(self._base_url, 'invoices', invoice_id, 'generate-qr-code')
        body = json.dumps({ 'width': width, 'height': height, 'action': action })
        response = self._session.post(url, body)

        if response.status_code != 200:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)        

    def record_invoice_payment(self, invoice_id: str, payment_detail: PaymentDetail) -> PaypalApiResponse:
        """ Calls the API to record a payment for the invoice.
            If no payment is due, the invoice is marked as PAID. 
            Otherwise, the invoice is marked as PARTIALLY PAID.
        
        Arguments:
            invoice_id {str} -- id of the invoice
            payment_detail {PaymentDetail} -- payment details
        
        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """
        url = parse_url(self._base_url, 'invoices', invoice_id, 'payments')
        body = json.dumps({ k : v for k,v in payment_detail.to_dict().items() if v != None })

        response = self._session.post(url, body)

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)        

    def delete_external_payment(self, invoice_id: str, transaction_id: str)  -> PaypalApiResponse:
        """ Deletes an external payment, by invoice ID and transaction ID.
        
        Arguments:
            invoice_id {str} -- id of the invoice
            transaction_id {str} --  The ID of the external payment transaction to delete.
        
        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """
        response = self._session.delete(parse_url(self._base_url, 'invoices', invoice_id, 'payments', transaction_id))

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def record_invoice_refund(self, invoice_id: str, refund_detail: RefundDetail)  -> PaypalApiResponse:
        """Calls the API to record a refund for the invoice.
           If all payments are refunded, the invoice is marked as REFUNDED. 
           Otherwise, the invoice is marked as PARTIALLY REFUNDED.
        
        Arguments:
            invoice_id {str} -- id of the invoice
            payment_detail {RefundDetail} -- payment details
        
        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """
        url = parse_url(self._base_url, 'invoices', invoice_id, 'refunds')
        body = json.dumps({ k : v for k,v in refund_detail.to_dict().items() if v != None })

        response = self._session.post(url, body)

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)        
        
    def delete_external_refund(self, invoice_id: str, transaction_id: str)  -> PaypalApiResponse:
        """  Deletes an external refund, by invoice ID and transaction ID.
        
        Arguments:
            invoice_id {str} -- id of the invoice
            transaction_id {str} --  The ID of the external payment transaction to delete.
        
        Returns:
            PaypalApiResponse -- Api operation response status containing the response
        """
        response = self._session.delete(parse_url(self._base_url, 'invoices', invoice_id, 'refunds', transaction_id))

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def send_invoice_reminder(
        self, invoice_id: str, subject: str, note: str, send_to_invoicer: bool,
        send_to_recipient: bool, additional_recipients: List[str] = []) -> PaypalApiResponse[Invoice]:
        """Calls the paypal API to sends a reminder to the payer about an invoice, by ID. 
           In the JSON request body, include a notification object that defines the subject 
           of the reminder and other details. 
        
        Arguments:
            invoice_id {str} -- invoice id
            subject {str} --  The subject of the email that is sent as a notification to the recipient.
            note {str} -- A note to the payer.
            send_to_invoicer {bool} -- Indicates whether to send a copy of the email to the merchant.
            send_to_recipient {bool} -- Indicates whether to send a copy of the email to the recipient.
            additional_recipients {List[str]} -- An array of one or more CC: emails to which notifications are sent.
             If ignored, a notification is sent to all CC: email addresses that are part of the invoice.

        Returns:
            PaypalApiResponse -- Api operation response status with parsed objects within
        """
        url = parse_url(self._base_url, 'invoices', invoice_id, 'remind')

        body = json.dumps({
            'subject': subject,
            'note': note,
            'send_to_invoicer': send_to_invoicer,
            'send_to_recipient': send_to_recipient
        })

        if additional_recipients:
            body['additional_recipients'] = [{'email_address': x} for x in additional_recipients]

        response = self._session.post(url, body)

        if response.status_code != 204:
            return PaypalApiResponse.error(response)
        
        return PaypalApiResponse.success(response)

    def send_invoice(
        self, invoice_id: str, subject: str, note: str, send_to_invoicer: bool,
        send_to_recipient: bool, additional_recipients: List[str] = [], paypal_request_id: str = None) -> PaypalApiResponse[Invoice]:
        """Calls the paypal API to send or schedule an invoice, by ID, to be sent to a customer. The action depends on the invoice issue date:
            If the invoice issue date is current or in the past, sends the invoice immediately.
            If the invoice issue date is in the future, schedules the invoice to be sent on that date.
            To suppress the merchant's email notification, set the send_to_invoicer body parameter to false. 
            To send the invoice through a share link and not through PayPal, set the send_to_recipient parameter to false in the notification object. 
            The send_to_recipient parameter does not apply to a future issue date because the invoice is scheduled to be sent through PayPal on that date.
        
        Arguments:
            invoice_id {str} -- invoice id
            subject {str} --  The subject of the email that is sent as a notification to the recipient.
            note {str} -- A note to the payer.
            send_to_invoicer {bool} -- Indicates whether to send a copy of the email to the merchant.
            send_to_recipient {bool} -- Indicates whether to send a copy of the email to the recipient.

        Keyword Arguments:
            paypal_request_id {str} -- Paypal request id for idempotence. (default: {None})
            additional_recipients {List[str]} -- An array of one or more CC: emails to which notifications are sent. (default: {[]})

        Returns:
            PaypalApiResponse -- Api operation response status with parsed objects within
        """
        response = None
        url = parse_url(self._base_url, 'invoices', invoice_id, 'send')

        body = json.dumps({
            'subject': subject,
            'note': note,
            'send_to_invoicer': send_to_invoicer,
            'send_to_recipient': send_to_recipient
        })

        if additional_recipients:
            body['additional_recipients'] = [{'email_address': x} for x in additional_recipients]

        if paypal_request_id:
            response = self._session.post(url, body, headers = { 'PayPal-Request-Id' : paypal_request_id })
        else:
            response = self._session.post(url, body)

        if response.status_code // 100 != 2:
            return PaypalApiResponse.error(response)

        if response.status_code == 200:
            return PaypalApiResponse.success(response, Invoice.serialize_from_json(response.json()))

        return PaypalApiResponse.success(response)

    def search_invoices(self, page: int, page_size: int, total_required: bool, search: InvoiceSearchRequest) -> PaypalPage[Invoice]:
        """Searches for and lists invoices that match search criteria. 
           If you pass multiple criteria, the response lists invoices 
           that match all criteria.

        Arguments:
            page {int} -- current page
            page_size {int} -- page size
            total_required {bool} -- total count required
            search {InvoiceSearchRequest} -- search criteria to be matched
        
        Returns:
            PaypalPage[Invoice] -- The paged elements in paypal API paged response 
        """
        query_params = { page: page, page_size: page_size, total_required: total_required }
        
        response = self._session.post(
            parse_url(self._base_url, 'search-invoices'), 
            json.dumps(search.to_dict()), params = query_params
        )

        if response.status_code != 200:
            return PaypalPage.error(response)
        
        json_response = response.json()
        items = [ Invoice.serialize_from_json(x) for x in json_response['items']]
        links = [ ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in json_response['links'] ]

        return PaypalPage.success(response, json_response.get('total_items'), json_response.get('total_pages'), items, links)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)
    
class InvoiceTemplateClient(ClientBase):
    """Client for invoice template resources
    """

    def __init__(self, base_url: str, session: PayPalSession):
            super().__init__(base_url, session)

    def create_template(self, template: Template, *, headers: dict = dict()) -> PaypalApiResponse:
        """Calls the Paypal API to create a template.
        
        Arguments:
            template {Template} -- object with all the required info.
        
        Returns:
            PaypalApiResponse[Template] -- The paypal API response
        """
        response = self._session.post(self._base_url, json.dumps(template.to_dict()), headers)

        if response.status_code != 200:
            return PaypalPage.error(response)

        parsed_data = Template.serialize_from_json(response.json()) if response.json() else None

        return PaypalApiResponse.success(response, parsed_data)
    
    def list_templates(self, page: int = 1, page_size: int = 20, fields: InvoiceListRequestField = InvoiceListRequestField.ALL) -> PaypalApiBulkResponse[Invoice]:
        """ Calls the paypal API to lists merchant-created templates with associated details. The associated details include 
            the emails, addresses, and phone numbers from the user's PayPal profile.
            The user can select which values to show in the business information section of their template.
        
        Keyword Arguments:
            page {int} -- current page index (default: {1})
            page_size {int} -- page size (how many elements per page) (default: {20})
            fields {InvoiceListRequestField} -- The fields to return in the response (default: {InvoiceListRequestField.NONE})
        
        Returns:
            PaypalApiBulkResponse[Invoice] -- A list of invoices
        """
        # TODO: There might be an error in the docs, the most logical response would be 
        # a page of templates, with the desired fields. This response must be tested.
        params = { 'page': page, 'page_size': page_size, 'fields': fields.name.lower() }

        response = self._session.get(self._base_url, params)

        if response.status_code != 200:
            return PaypalApiBulkResponse.error(response)

        return PaypalApiBulkResponse.success(response, Invoice.serialize_from_json(response.json()))
        
    def delete_template(self, template_id: str) -> PaypalApiResponse:
        """Deletes a template by id
        
        Arguments:
            template_id {str} -- The id of the template to be deleted
        
        Returns:
            PaypalApiResponse -- API Response.
        """
        response = self._session.delete(parse_url(self._base_url, template_id))
        return PaypalApiResponse.success(response) if response.status_code == 204 else PaypalApiResponse.error(response)
    
    def update_template(self, template_id: str, template: Template) -> PaypalApiResponse[Template]:
        """Fully updates a template
        
        Arguments:
            template_id {str} -- the template id
            template {Template} -- template info with updates
        
        Returns:
            PaypalApiResponse[Template] -- A response that might contain the template
        """
        response = self._session.put(parse_url(self._base_url, template_id), json.dumps(template.to_dict()))

        # TODO: The docs are inconsistent on this call response. Test manually
        if response.status_code // 100 != 2:
            return PaypalApiResponse.error(response)
        
        parsed_data = Template.serialize_from_json(response.json()) if response.json() else None
        return PaypalApiResponse.success(response, parsed_data)
    
    def show_template_details(self, template_id: str) -> PaypalApiResponse[Template]:
        """Shows the details for a template by it's id
        
        Arguments:
            template_id {str} -- The template id
        
        Returns:
            PaypalApiResponse[Template] -- Response status with a Template object
        """
        response = self._session.get(parse_url(self._base_url, template_id))

        if response.status_code != 200:
            return PaypalApiResponse.error(response)

        return PaypalApiResponse.success(response, Invoice.serialize_from_json(response.json()))

    @classmethod
    def for_session(cls: I, session: PayPalSession) -> I:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = parse_url(_LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL, 'templates')
        return cls(base_url, session)


"""
    Module with all Paypal Sync (transactions) related entities 
"""
from enum import Enum
from datetime import datetime
from typing import NamedTuple, Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T,
    Money,
    PaypalName,
    ActionLink,
    ResponseType,
    PayPalEntity,
    PaypalPhoneDetail,
    PaypalPortableAddress
)

# Transaction statuses string definitions
# See the enumeration for details.
_TRANSACTION_STATUSES = [ 'D', 'F', 'P', 'S', 'V' ]

class TransactionStatus(Enum):
    # PayPal or merchant rules denied the transaction. ('D')
    DENIED = 1
    # The original recipient partially refunded the transaction. ('F')
    PARTIALLY_REFUNDED = 2
    # The transaction is pending. The transaction was created but waits for another payment process to complete, 
    # such as an ACH transaction, before the status changes to S. ('P')
    PENDING = 3
    # The transaction successfully completed without a denial and after any pending statuses. ('S')
    SUCCESSFUL = 4
    # A successful transaction was reversed and funds were refunded to the original sender. ('V')
    SUCCESSFUL_REVERSAL = 5

    def as_string_parameter(self) -> str:
        return _TRANSACTION_STATUSES[self.value - 1]

# Transaction status mapping for enum conversion
_TRANSACTION_STATUS_MAPPING = {
    'D': TransactionStatus.DENIED,
    'P': TransactionStatus.PENDING,
    'S': TransactionStatus.SUCCESSFUL,
    'F': TransactionStatus.PARTIALLY_REFUNDED,
    'V': TransactionStatus.SUCCESSFUL_REVERSAL
}

# Mapping dictionary for different transaction event types.
_EVENT_TYPE_MAPPINGS = {
    'T00': 'PayPal account-to-PayPal account payment',
    'T01': 'Non-payment-related fees',
    'T02': 'Currency conversion',
    'T03': 'Bank deposit into PayPal account',
    'T04': 'Bank withdrawal from PayPal account',
    'T05': 'Debit card',
    'T06': 'Credit card withdrawal',
    'T07': 'Credit card deposit',
    'T08': 'Bonus',
    'T09': 'Incentive',
    'T10': 'Bill pay',
    'T11': 'Reversal',
    'T12': 'Adjustment',
    'T13': 'Authorization',
    'T14': 'Dividend',
    'T15': 'Hold for dispute or other investigation',
    'T16': 'Buyer credit deposit',
    'T17': 'Non-bank withdrawal',
    'T18': 'Buyer credit withdrawal',
    'T19': 'Account correction',
    'T20': 'Funds transfer from PayPal account to another',
    'T21': 'Reserves and releases',
    'T22': 'Transfers',
    'T30': 'Generic instrument and Open Wallet',
    'T50': 'Collections and disbursements',
    'T97': 'Payables and receivables',
    'T98': 'Display only transaction',
    'T99': 'Other',
}

# Mapping dictionary for different transaction event descriptions.
_EVENT_DESC_MAPPINGS = {
    # PayPal account-to-PayPal account payment
    'T0000': 'General: received payment of a type not belonging to the other T00nn categories.',
    'T0001': 'MassPay payment.',
    'T0002': 'Subscription payment. Either payment sent or payment received.',
    'T0003': 'Pre-approved payment (BillUser API). Either sent or received.',
    'T0004': 'eBay auction payment.',
    'T0005': 'Direct payment API.',
    'T0006': 'PayPal Checkout APIs.',
    'T0007': 'Website payments standard payment.',
    'T0008': 'Postage payment to carrier.',
    'T0009': 'Gift certificate payment. Purchase of gift certificate.',
    'T0010': 'Third-party auction payment.',
    'T0011': 'Mobile payment, made through a mobile phone.',
    'T0012': 'Virtual terminal payment.',
    'T0013': 'Donation payment.',
    'T0014': 'Rebate payments.',
    'T0015': 'Third-party payout.',
    'T0016': 'Third-party recoupment.',
    'T0017': 'Store-to-store transfers.',
    'T0018': 'PayPal Here payment.',
    'T0019': 'Generic instrument-funded payment.',

    # Non-payment-related fees
    'T0100': 'General non-payment fee of a type not belonging to the other  T01nn categories.',
    'T0101': 'Website payments. Pro account monthly fee.',
    'T0102': 'Foreign bank withdrawal fee.',
    'T0103': 'WorldLink check withdrawal fee.',
    'T0104': 'Mass payment batch fee.',
    'T0105': 'Check withdrawal.',
    'T0106': 'Chargeback processing fee. Chargeback processing, or handling, fee. Charged against account that received a chargeback for a previous transaction.',
    'T0107': 'Payment fee.',
    'T0108': 'ATM withdrawal.',
    'T0109': 'Auto-sweep from account.',
    'T0110': 'International credit card withdrawal.',
    'T0111': 'Warranty fee for warranty purchase.',
    'T0112': 'Gift certificate expiration fee.',
    'T0113': 'Partner fee.',

    # Currency conversion
    'T0200': 'General currency conversion. Transfer of funds from one currency balance to a different currency balance or Withdrawal of funds from one currency balance that is covered by funds from a different currency balance.',
    'T0201': 'User-initiated currency conversion.',
    'T0202': 'Currency conversion required to cover negative balance. PayPal-system generated.',
    
    # Bank deposit into PayPal account
    'T0300': 'General funding of PayPal account. Deposit to PayPal balance from a bank account.',
    'T0301': 'PayPal balance manager funding of PayPal account.(PayPal-system generated).',
    'T0302': 'ACH funding for funds recovery from account balance.',
    'T0303': 'Electronic funds transfer (EFT) (German banking).',

    # Bank withdrawal from PayPal account
    'T0400': 'General withdrawal from PayPal account. Settlement withdrawal or user-initiated.',
    'T0401': 'AutoSweep.',

    # Debit card
    'T0500': 'General PayPal debit card transaction. Requires a PayPal debit card associated with the PayPal account.',
    'T0501': 'Virtual PayPal debit card transaction.',
    'T0502': 'PayPal debit card withdrawal to ATM.',
    'T0503': 'Hidden virtual PayPal debit card transaction.',
    'T0504': 'PayPal debit card cash advance.',
    'T0505': 'PayPal debit authorization.',

    # Credit card withdrawal
    'T0600': 'General credit card withdrawal. Reversal of purchase with a credit card. Seen only in PayPal account of the credit card owner.',

    # Credit card deposit
    'T0700': 'General credit card deposit. Purchase with a credit card.',
    'T0701': 'Credit card deposit for negative PayPal account balance.',

    # Bonus
    'T0800': 'General bonus of a type not belonging to the other  T08nn categories.',
    'T0801': 'Debit card cash back bonus. Requires a PayPal debit card associated with the PayPal account.',
    'T0802': 'Merchant referral account bonus. Must have created a merchant referral bonus link.',
    'T0803': 'Balance manager account bonus.',
    'T0804': 'PayPal buyer warranty bonus.',
    'T0805': 'PayPal protection bonus, payout for PayPal buyer protection, payout for full protection with PayPal buyer credit.',
    'T0806': 'Bonus for first ACH use.',
    'T0807': 'Credit card security charge refund.',
    'T0808': 'Credit card cash back bonus.',
    
    # Incentive
    'T0900': 'General incentive or certificate redemption.',
    'T0901': 'Gift certificate redemption.',
    'T0902': 'Points incentive redemption.',
    'T0903': 'Coupon redemption.',
    'T0904': 'eBay loyalty incentive.',
    'T0905': 'Offers used as funding source.',

    # Bill pay
    'T1000': 'Bill pay transaction.',

    # Reversal
    'T1100': 'General reversal of a type not belonging to the other  T11nn categories.',
    'T1101': 'Reversal of ACH withdrawal transaction. Reversal of a withdrawal from PayPal balance to a bank account.',
    'T1102': 'Reversal of debit card transaction. Reversal of a debit card payment. Requires a PayPal debit card.',
    'T1103': 'Reversal of points usage.',
    'T1104': 'Reversal of ACH deposit.',
    'T1105': 'Reversal of general account hold.',
    'T1106': 'Payment reversal, initiated by PayPal. Completion of a chargeback.',
    'T1107': 'Payment refund, initiated by merchant.',
    'T1108': 'Fee reversal.',
    'T1109': 'Fee refund.',
    'T1110': 'Hold for dispute investigation (T15nn). To cover possible chargeback.',
    'T1111': 'Cancellation of hold for dispute resolution. Cancellation of temporary hold to cover possible chargeback.',
    'T1112': 'MAM reversal.',
    'T1113': 'Non-reference credit payment.',
    'T1114': 'MassPay reversal transaction.',
    'T1115': 'MassPay refund transaction.',
    'T1116': 'Instant payment review (IPR) reversal.',
    'T1117': 'Rebate or cash back reversal.',
    'T1118': 'Generic instrument/Open Wallet reversals (seller side).',
    'T1119': 'Generic instrument/Open Wallet reversals (buyer side).',

    # Adjustment
    'T1200': 'General account adjustment.',
    'T1201': 'Chargeback.',
    'T1202': 'Chargeback reversal.',
    'T1203': 'Charge-off adjustment.',
    'T1204': 'Incentive adjustment.',
    'T1205': 'Reimbursement of chargeback.',
    'T1207': 'Chargeback re-presentment rejection. Adjustment to PayPal account for reversal of payment based on rejection of the re-presentment for a chargeback in the PayPal system.',
    'T1208': 'Chargeback cancellation. Adjustment to PayPal account for reversal of chargeback reversal based on cancellation of a chargeback in the PayPal system.',

    # Authorization
    'T1300': 'General authorization.',
    'T1301': 'Reauthorization.',
    'T1302': 'Void of authorization.',

    # Dividend
    'T1400': 'General dividend.',

    # Hold for dispute or other investigation
    'T1500': 'General temporary hold of a type not belonging to the other  T15nn categories.',
    'T1501': 'Account hold for open authorization.',
    'T1502': 'Account hold for ACH deposit.',
    'T1503': 'Temporary hold on available balance.',

    # Buyer credit deposit
    'T1600': 'PayPal buyer credit payment funding. Must have signed up for buyer credit.',
    'T1601': 'BML credit. Transfer from BML.',
    'T1602': 'Buyer credit payment.',
    'T1603': 'Buyer credit payment withdrawal. Transfer to BML.',

    # Non-bank withdrawal
    'T1700': 'General withdrawal to non-bank institution.',
    'T1701': 'WorldLink withdrawal.',

    # Buyer credit withdrawal
    'T1800': 'General buyer credit payment. Must have signed up for buyer credit.',
    'T1801': 'BML withdrawal. Transfer to BML.',

    # Account correction
    'T1900': 'General adjustment without business-related event.',

    # Funds transfer from PayPal account to another
    'T2000': 'General intra-account transfer.',
    'T2001': 'Settlement consolidation.',
    'T2002': 'Transfer of funds from payable.',
    'T2003': 'Transfer to external GL entity.',

    # Reserves and releases
    'T2101': 'General hold.',
    'T2102': 'General hold release.',
    'T2103': 'Reserve hold. PayPal holds (n) of funds for (d) days as a condition for processing for you. This amount is part of your total balance and is not available for withdrawal.',
    'T2104': 'Reserve release. A reserve, when released, is now available for withdrawal.',
    'T2105': 'Payment review hold. A payment review hold is used to protect you against unauthorized fraud loss and for seller protection. While a transaction is on payment review hold, we recommend that you do not ship the item. We are reviewing this transaction for up to 24 hours.',
    'T2106': 'Payment review release. A payment review hold, when released, is now available for withdrawal.',
    'T2107': 'Payment hold. Payment holds are funds that belong to you that we set aside, such as a security deposit. This amount is part of your total balance and is not available for withdrawal.',
    'T2108': 'Payment hold release. A payment hold, when released, is now available for withdrawal.',
    'T2109': 'Gift certificate purchase. When a gift certificate is purchased by your buyer, that amount is placed on hold.',
    'T2110': 'Gift certificate redemption. A gift certificate, when redeemed, is available for withdrawal.',
    'T2111': 'Funds not yet available. While you establish a successful sales history on eBay, funds from eBay sales are usually available in 21 or fewer days, based on how you ship the order. You might get your funds faster if you print your shipping labels on eBay or PayPal, upload tracking information, or mark the item as shipped on eBay. If your buyer reports a problem with the sale, it might take longer to get funds.',
    'T2112': 'Funds available. These funds are available for use.',
    'T2113': 'Blocked payments.',

    # Transfers
    'T2201': 'Transfer to and from a credit-card-funded restricted balance.',

    # Generic instrument and Open Wallet
    'T3000': 'Generic instrument/Open Wallet transaction.',

    # Collections and disbursements
    'T5000': 'Deferred disbursement. Funds collected for disbursement.',
    'T5001': 'Delayed disbursement. Funds disbursed.',

    # Payables and receivables
    'T9700': 'Account receivable for shipping.',
    'T9701': 'Funds payable. PayPal-provided funds that must be paid back.',
    'T9702': 'Funds receivable. PayPal-provided funds that are being paid back.',

    # Display only transaction
    'T9800': 'Display only transaction.',

    # Other
    'T9900': 'Other.'
}

class ReferenceIdType(Enum):
    # An order ID. (ODR)
    ORDER_ID = 1
    # A transaction ID. (TXN)
    TRANSACTION_ID = 2
    # A subscription ID. (SUB)
    SUBSCRIPTION_ID = 3 
    # A pre-approved payment ID. (PAP)
    PREAPPROVED_PAYMENT_ID = 4

_REFERENCE_ID_MAPPINGS = {
    'ODR': ReferenceIdType.ORDER_ID,
    'TXN': ReferenceIdType.TRANSACTION_ID,
    'SUB': ReferenceIdType.SUBSCRIPTION_ID,
    'PAP': ReferenceIdType.PREAPPROVED_PAYMENT_ID
}

class ProtectionElegibility(Enum):
    ELEGIBLE = 1
    NOT_ELEGIBLE = 2
    PARTIALLY_ELEGIBLE = 3

class PayerStatusProperty(Enum):
    VERIFIED = 1
    UNVERIFIED = 2

class TransactionInfo(PayPalEntity):
    """Transaction info object representation
    """

    _MONEY_TYPES = {
        'transaction_amount', 'fee_amount', 'discount_amount', 'insurance_amount',
        'sales_tax_amount', 'shipping_amount', 'shipping_discount_amount',
        'shipping_tax_amount', 'other_amount', 'tip_amount', 'ending_balance',
        'available_balance', 'credit_transactional_fee', 'credit_promotional_fee'
    }

    def __init__(
        self, paypal_account_id: str = None, transaction_id: str = None,
        paypal_reference_id: str = None, paypal_reference_id_type: str = None, 
        transaction_event_code: str = None, transaction_amount: Money = None, fee_amount: Money = None, 
        discount_amount: Money = None, insurance_amount: Money = None, sales_tax_amount: Money = None, 
        shipping_amount: Money = None, shipping_discount_amount: Money = None, shipping_tax_amount: Money = None, 
        other_amount: Money = None, tip_amount: Money = None, transaction_status: str = None, transaction_subject: str = None, 
        transaction_note: str = None, payment_tracking_id: str = None, bank_reference_id: str = None, ending_balance: Money = None, 
        available_balance: Money = None, invoice_id: str = None, custom_field: str = None, protection_eligibility: str = None, 
        credit_term: str = None, credit_transactional_fee: Money = None, credit_promotional_fee: Money = None, 
        annual_percentage_rate: str = None, payment_method_type: str = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.fee_amount = fee_amount
        self.tip_amount = tip_amount
        self.invoice_id = invoice_id
        self.credit_term = credit_term
        self.other_amount = other_amount
        self.custom_field = custom_field
        self.ending_balance = ending_balance
        self.transaction_id = transaction_id
        self.discount_amount = discount_amount
        self.shipping_amount = shipping_amount
        self.insurance_amount = insurance_amount
        self.sales_tax_amount = sales_tax_amount
        self.transaction_note = transaction_note
        self.bank_reference_id = bank_reference_id
        self.paypal_account_id = paypal_account_id
        self.available_balance = available_balance
        self.transaction_amount = transaction_amount
        self.transaction_status = transaction_status
        self.transaction_subject = transaction_subject
        self.paypal_reference_id = paypal_reference_id
        self.shipping_tax_amount = shipping_tax_amount
        self.payment_tracking_id = payment_tracking_id
        self.payment_method_type = payment_method_type
        self.protection_eligibility = protection_eligibility
        self.transaction_event_code = transaction_event_code
        self.credit_promotional_fee = credit_promotional_fee
        self.annual_percentage_rate = annual_percentage_rate
        self.paypal_reference_id_type = paypal_reference_id_type
        self.shipping_discount_amount = shipping_discount_amount
        self.credit_transactional_fee = credit_transactional_fee
        self._transaction_updated_date = self._json_response.get('transaction_updated_date', kwargs.get('transaction_updated_date'))
        self._transaction_initiation_date = self._json_response.get('transaction_initiation_date', kwargs.get('transaction_initiation_date'))
    
    @property
    def event_type(self) -> str:
        """Property for the event type based on the code
        
        Returns:
             -- String description of the event type or None
        """
        try:
            return _EVENT_TYPE_MAPPINGS.get(self.transaction_event_code[0:3])
        except:
            return None

    @property
    def event_desc(self) -> str:
        """Property for the event description based on the code
        
        Returns:
             -- String description of the event type or None
        """
        return _EVENT_DESC_MAPPINGS.get(self.transaction_event_code)

    @property
    def transaction_status_enum(self) -> TransactionStatus:
        """Property for the enumeration of this transaction ()
        """
        return _TRANSACTION_STATUS_MAPPING.get(self.transaction_status)

    @property
    def paypal_reference_id_type_enum(self) -> ReferenceIdType:
        """Property for the enumeration of this transaction ()
        """
        return _REFERENCE_ID_MAPPINGS.get(self.paypal_reference_id_type)

    @property
    def protection_eligibility_enum(self) -> ProtectionElegibility:
        """Property for the enumeration of this transaction ()
        """
        try:
            return ProtectionElegibility(int(self.protection_eligibility))
        except:
            return None

    @property
    def transaction_updated_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._transaction_updated_date) if self._transaction_updated_date else None
        except:
            return None

    @property
    def transaction_initiation_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._transaction_initiation_date) if self._transaction_initiation_date else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = {**json_data , **{ k : Money.serialize_from_json(v) for k, v in json_data.items() if k in cls._MONEY_TYPES }}
        return cls(**args, json_response = json_data, response_type = response_type)

class PayerInfo(PayPalEntity):
    """Payer info object representation
    """
    
    _PAYPAL_ENTITY_TYPES = { 'payer_name': PaypalName, 'address': PaypalPortableAddress, 'phone_number': PaypalPhoneDetail }

    def __init__(
        self, account_id: str = None, email_address: str = None, 
        phone_number: PaypalPhoneDetail = None, address_status: str = None, 
        payer_status: str = None, payer_name: PaypalName = None, country_code: str = None,
        address: PaypalPortableAddress = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.address = address
        self.payer_name = payer_name
        self.account_id = account_id
        self.country_code = country_code
        self.payer_status = payer_status
        self.phone_number = phone_number
        self.email_address = email_address
        self.address_status = address_status

    @property
    def payer_status_enum(self) -> PayerStatusProperty:
        """Payer status enumeration
        
        Returns:
            PayerStatusProperty -- A descriptive enumeration for the payers status.
        """
        stat = None
        if self.payer_status:
            stat = PayerStatusProperty.VERIFIED if self.payer_status == 'Y' else PayerStatusProperty.UNVERIFIED
        return stat

    @property
    def address_status_enum(self) -> PayerStatusProperty:
        """Address status enumeration
        
        Returns:
            PayerStatusProperty -- A descriptive enumeration for the payers address status.
        """
        stat = None
        if self.address_status:
            stat = PayerStatusProperty.VERIFIED if self.address_status == 'Y' else PayerStatusProperty.UNVERIFIED
        return stat

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        special_types = { k : cls._PAYPAL_ENTITY_TYPES[k].serialize_from_json(v) for k,v in json_data.items() if k in cls._PAYPAL_ENTITY_TYPES.keys() }
        args = { **json_data, **special_types }
        return cls(**args, json_response = json_data, response_type = response_type)

class ShippingInfo(PayPalEntity):
    """Shipping info object representation
    """
    _ADDRESS_TYPES = {'address', 'secondary_shipping_address'}

    def __init__(
        self, name: str = None, method: str = None, 
        address: PaypalPortableAddress = None, 
        secondary_shipping_address: PaypalPortableAddress = None,
        **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.method = method
        self.address = address
        self.secondary_shipping_address = secondary_shipping_address
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = {**json_data , **{ k : PaypalPortableAddress.serialize_from_json(v) for k, v in json_data.items() if k in cls._ADDRESS_TYPES }}
        return cls(**args, json_response = json_data, response_type = response_type)

class TaxAmount(PayPalEntity):
    """Tax amount object representation.
    """

    def __init__(self, tax_amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.tax_amount = tax_amount
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(Money.serialize_from_json(json_data['tax_amount']), json_response = json_data, response_type = response_type)

class CheckoutOption(PaypalName):
    """Checkout option object representation.
    """

    def __init__(self, checkout_option_name: str, checkout_option_value: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.checkout_option_name = checkout_option_name
        self.checkout_option_value = checkout_option_value

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class ItemDetail(PayPalEntity):
    """Item detail object representation.
    """

    _MONEY_TYPES = {
        'item_unit_price', 'item_amount', 'discount_amount', 'adjustment_amount', 
        'gift_wrap_amount', 'basic_shipping_amount', 'extra_shipping_amount', 
        'handling_amount', 'insurance_amount', 'total_item_amount'
    }

    _PAYPAL_TYPE_ARRAYS = {
        'tax_amounts': TaxAmount,
        'checkout_options': CheckoutOption
    }

    def __init__(
        self, item_code: str = None, item_name: str = None, item_description: str = None, 
        item_options: str = None, item_quantity: str = None, item_unit_price: Money = None, 
        item_amount: Money = None, discount_amount: Money = None, adjustment_amount: Money = None, 
        gift_wrap_amount: Money = None, tax_percentage: str = None, tax_amounts: List[TaxAmount] = [], 
        basic_shipping_amount: Money = None, extra_shipping_amount: Money = None, handling_amount: Money = None, 
        insurance_amount: Money = None, total_item_amount: Money = None, invoice_number: str = None, 
        checkout_options: List[CheckoutOption] = [], **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.item_code = item_code
        self.item_name = item_name
        self.item_description = item_description
        self.item_options = item_options
        self.item_quantity = item_quantity
        self.item_unit_price = item_unit_price
        self.item_amount = item_amount
        self.discount_amount = discount_amount
        self.adjustment_amount = adjustment_amount
        self.gift_wrap_amount = gift_wrap_amount
        self.tax_percentage = tax_percentage
        self.tax_amounts = tax_amounts or []
        self.basic_shipping_amount = basic_shipping_amount
        self.extra_shipping_amount = extra_shipping_amount
        self.handling_amount = handling_amount
        self.insurance_amount = insurance_amount
        self.total_item_amount = total_item_amount
        self.invoice_number = invoice_number
        self.checkout_options = checkout_options or []
        
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = {
            # Regular (primitive) arguments
            **json_data,
            # Serialized Money arguments
            **{ k : Money.serialize_from_json(v) for k, v in json_data.items() if k in cls._MONEY_TYPES },
            # Serialized array arguments
            **{ k : [cls._PAYPAL_TYPE_ARRAYS[k].serialize_from_json(v) for v in json_data[k]] for k in json_data.keys() if k in cls._PAYPAL_TYPE_ARRAYS }
        }
        return cls(**args, json_response = json_data, response_type = response_type)

class CartInfo(PayPalEntity):
    """Cart info object representation
    """
    
    def __init__(self, tax_inclusive: bool = None, paypal_invoice_id: str = None, item_details: List[ItemDetail] = [], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.tax_inclusive = tax_inclusive
        self.item_details = item_details or []
        self.paypal_invoice_id = paypal_invoice_id
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        items = []
        if 'item_details' in json_data.keys():
            items = [ItemDetail.serialize_from_json(v) for v in json_data['item_details']]        
        return cls(json_data.get('tax_inclusive'), json_data.get('tax_inclusive'), items)

class StoreInfo(PayPalEntity):
    """Store info object representation.
    """

    def __init__(self, store_id: str = None, terminal_id: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.store_id = store_id
        self.terminal_id = terminal_id
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class AuctionInfo(PayPalEntity):
    """Auction info object representation
    """

    def __init__(
        self, auction_site: str = None, auction_item_site: str = None, 
        auction_buyer_id: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.auction_site = auction_site
        self.auction_item_site = auction_item_site
        self.auction_buyer_id = auction_buyer_id
        self._auction_closing_date = self._json_response.get('auction_closing_date', kwargs.get('auction_closing_date'))
    
    @property
    def auction_closing_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._auction_closing_date) if self._auction_closing_date else None
        except:
            return None
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class IncentiveDetail(PayPalEntity):
    """Incentive detail object representation
    """

    def __init__(
        self, incentive_type: str = None, incentive_code: str = None, 
        incentive_amount: Money = None, incentive_program_code: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.incentive_type = incentive_type
        self.incentive_code  = incentive_code 
        self.incentive_amount = incentive_amount
        self.incentive_program_code = incentive_program_code
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = { **json_data }
        if 'incentive_amount' in json_data.keys():
            args['incentive_amount'] = Money.serialize_from_json(json_data['incentive_amount'])
        return cls(**args, json_response = json_data, response_type = response_type)

class TransactionDetails(PayPalEntity):
    """Transaction details obj representation
    """

    _PAYPAL_TYPES = { 
        # The transaction information.
        'transaction_info' : TransactionInfo,

        # The payer information.
        'payer_info' : PayerInfo,

        # The shipping information.
        'shipping_info' : ShippingInfo,

        # The cart information.
        'cart_info' : CartInfo,

        # The store information.
        'store_info' : StoreInfo,

        # The auction information.   
        'auction_info' : AuctionInfo
    }

    def __init__(
        self, transaction_info: TransactionInfo, 
        payer_info: PayerInfo, shipping_info: ShippingInfo, 
        cart_info: CartInfo, store_info: StoreInfo, 
        auction_info: AuctionInfo, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.cart_info = cart_info
        self.store_info = store_info
        self.payer_info = payer_info
        self.auction_info = auction_info
        self.shipping_info = shipping_info
        self.transaction_info = transaction_info

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = { k : cls._PAYPAL_TYPES[k].serialize_from_json(v) for k,v in json_data.items() if k in cls._PAYPAL_TYPES and v }
        return cls(**args, json_response = json_data, response_type = response_type)

class TransactionResponse(PayPalEntity):
    """Transaction query response with the main info and details    
    """

    def __init__(self, account_number: str = None, 
    transaction_details: List[TransactionDetails] = [], **kwargs):
        self.account_number = account_number
        self.transaction_details = transaction_details or []
        self.page = self._json_response.get('page', kwargs.get('page'))
        self._end_date = self._json_response.get('end_date', kwargs.get('end_date'))
        self._start_date = self._json_response.get('start_date', kwargs.get('start_date'))
        self.total_pages = self._json_response.get('total_pages', kwargs.get('total_pages'))
        self.total_items = self._json_response.get('total_items', kwargs.get('total_items'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]
        self._last_refreshed_datetime = self._json_response.get('last_refreshed_datetime', kwargs.get('last_refreshed_datetime'))

    @property
    def end_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._end_date) if self._end_date else None
        except:
            return None

    @property
    def start_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._start_date) if self._start_date else None
        except:
            return None

    @property
    def last_refreshed_datetime(self) -> datetime:
        try:
            return dateutil.parser.parse(self._last_refreshed_datetime) if self._last_refreshed_datetime else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = { **json_data }
        if 'transaction_details' in json_data.keys():
            args['transaction_details'] = [ TransactionDetails.serialize_from_json(v) for v in json_data['transaction_details'] ]
        return cls(**args, json_response = json_data, response_type = response_type)



"""
    Product resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/catalog-products/v1/
"""

import json
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink
from pypaypal.entities.base import ResponseType, PaypalApiResponse, PaypalPage

from pypaypal.entities.product import Product, ProductType, ProductUpdateRequest

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL, 
    LEGACY_SANDBOX_API_BASE_URL 
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'catalogs/products')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'catalogs/products')

"""
    Supported product properties for product creation
"""
_PRODUCT_PROPERTIES = { 
    'name', 'product_type', 'id', 'category', 'home_url',
    'image_url', 'description', 'update_time', 'create_time'
}

T = TypeVar('T', bound = 'ProductsClient')


class ProductsClient(ClientBase):
    """Products resource group client class
    """

    def __init__(self, url: str, session: PayPalSession):
        """Client ctor
        
        Arguments:
            session {PayPalSession} -- The http session 
        """
        super().__init__(url, session)

    def create_product(self, product: Product, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Product]:
        """Performs an API call to create a new product
        
        Arguments:
            product_name {str} -- The product info
            product_type {ProductType} -- the response type (Minimal or Representation)
        """
        url = self._base_url
        headers = { 'Prefer': response_type.as_header_value() }
        
        body = product.to_dict()
        body['create_time'] = body.pop('_create_time', None)
        body['update_time'] = body.pop('_update_time', None)
        self._clean_dictionary(body, _PRODUCT_PROPERTIES)

        api_response = self._session.post(url, json.dumps(body), headers = headers)
        
        if api_response.status_code != 201:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response, Product.serialize_from_json(api_response.json(), response_type))

    def list_products(self, page_size: int, page: int) -> PaypalPage[Product]:
        """Performs an API call to get a page of products
        
        Arguments:
            page_size {int} -- Amount of elements in the page
            page {int} -- current page index
        
        Returns:
            PaypalPage[Product] -- A page with product elements
        """
        url = self._base_url
        params = { 'page_size': page_size, 'page': page, 'total_required': True }

        api_response = self._session.get(url, params)
        
        if api_response.status_code != 200:
            return PaypalPage(True, api_response, 0, 0, [], [])
        
        json_response = api_response.json()
        products = [ Product.serialize_from_json(x) for x in json_response['products']]
        links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in json_response['links']]

        return PaypalPage(False, api_response, json_response['total_items'], json_response['total_pages'], products, links)

    def list_products_from_page_link(self, page_link: ActionLink) -> PaypalPage[Product]:
        """Performs an API call to get a page of products
        
        Arguments:
            page_size {int} -- Amount of elements in the page
            page {int} -- current page index
        
        Returns:
            PaypalPage[Product] -- A page with product elements
        """
        return self._page_from_link(page_link, 'products', Product)

    def show_product_details(self, product_id: str, response_type = ResponseType.MINIMAL) -> PaypalApiResponse[Product]:
        """Calls the API to get the details for a given product
        
        Arguments:
            product_id {str} -- The product identifier
        
        Keyword Arguments:
            response_type {[type]} -- Response representation (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Product] -- Response status with data
        """
        url = parse_url(self._base_url, product_id)
        headers = { 'Prefer': response_type.as_header_value() }        
        api_response = self._session.get(url, None, headers = headers)

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Product.serialize_from_json(api_response.json(), response_type))
    
    def product_details_from_entity(self, product: Product, response_type: ResponseType=ResponseType.MINIMAL) -> PaypalApiResponse[Product]:
        """Calls the API to get the details for a given product
        
        Arguments:
            product {Product} -- The product entity with the HATEOAS link
        
        Keyword Arguments:
            response_type {[type]} -- Response representation (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Product] -- Response status with data
        """
        url = product.read_link
        headers = { 'Prefer': response_type.as_header_value() }        
        api_response = self._session.get(url, None, headers = headers)

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Product.serialize_from_json(api_response.json(), response_type))

    def update_product(self, product_id: str, updates: List[ProductUpdateRequest]) -> PaypalApiResponse:
        """Updates a product
        
        Arguments:
            product_id {str} -- [description]
            updates {List[ProductUpdateRequest]} -- [description]
        
        Returns:
            PaypalApiResponse -- [description]
        """
        url = parse_url(self._base_url, product_id)
        body = [ {'op': x.operation, 'path': x.path, 'value': x.value} for x in updates ]
        api_response = self._session.patch(url, json.dumps(body))

        return PaypalApiResponse(api_response.status_code != 204, api_response)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a product client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of TrackersClient with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)


"""
    Dispute resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/customer-disputes/v1/#disputes
"""

import json

from datetime import datetime
from typing import Type, TypeVar, List

from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication

from pypaypal.clients.base import ClientBase, ActionLink
from pypaypal.entities.dispute import Dispute, DisputeUpdateRequest, DisputeEvidence

from pypaypal.entities.base import ( 
    ResponseType, 
    PaypalApiResponse, 
    PaypalPage, 
    PaypalApiBulkResponse, 
    Money, 
    PaypalPortableAddress 
)

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL, 
    LEGACY_SANDBOX_API_BASE_URL 
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'disputes')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'disputes')

T = TypeVar('T', bound = 'DisputeClient')

class _SandboxExclusiveDisputeClient(ClientBase):
    """Disputes resource group client class for API calls 
       that should only be accesed in sandbox mode.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

class DisputeClient(ClientBase):
    """Disputes resource group client class
    """
    
    def __init__(self, url: str, session: PayPalSession):
        """Client ctor
        
        Arguments:
            session {PayPalSession} -- The http session 
        """
        super().__init__(url, session)
        self.sandbox_exclusive = None if session.session_mode.is_live() else _SandboxExclusiveDisputeClient(url, session)

    def list_disputes(self, start_time: datetime, page_size:int=2) -> PaypalPage[Dispute]:
        """Performs an API call to lists disputes
        
        Arguments:
            start_time {datetime} -- Start date
        
        Keyword Arguments:
            page_size {int} -- size of the page (default: {2})
        
        Returns:
            PaypalPage[Dispute] -- Page with a lists of disputes
        """
        url = self._base_url
        params = { 'page_size' : page_size, 'start_time': start_time.strftime('%Y-%m-%dT%H:%M:%S.000Z') }

        api_response = self._session.get(url, params)

        if api_response.status_code != 200:
            return PaypalPage(True, api_response, 0, 0, [], [])
        
        json_response = api_response.json()
        items = [ Dispute.serialize_from_json(x) for x in json_response['items']]
        links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in json_response['links']]

        return PaypalPage(False, api_response, json_response.get('total_items'), json_response.get('total_pages'), items, links)

    def list_disputes_from_link(self, link: ActionLink) -> PaypalPage[Dispute]:
        """Performs an API call to lists disputes inside a page
        
        Arguments:
            link {ActionLink} -- page link
                
        Returns:
            PaypalPage[Dispute] -- Page with a lists of disputes
        """
        return self._page_from_link(link, 'items', Dispute)
    
    def partial_dispute_update(self, dispute_id: str, update_request: DisputeUpdateRequest) -> PaypalApiResponse[Dispute]:
        """Calls the Paypal API to  Partially update a dispute, by ID.            
            Right now it's only possible to update the communication_detail value.
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            update_request {DisputeUpdateRequest} -- Update request with the details
        
        Returns:
            PaypalApiResponse[Dispute] -- Operation response status
        """
        url = parse_url(self._base_url, dispute_id)
        body = { 'op': update_request.operation, 'path': update_request.path, 'value': update_request.value }

        api_response = self._session.patch(url, json.dumps(body))

        return PaypalApiResponse(api_response.status_code != 204, api_response)

    def show_dispute_details(self, dispute_id: str, response_type: ResponseType=ResponseType.MINIMAL) -> PaypalApiResponse[Dispute]:
        """Performs an API call to show the details for a dispute, by ID. 
        
        Arguments:
            dispute_id {str} -- Dispute identifier
        
        Keyword Arguments:
            response_type {ResponseType} -- The preferred response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Dispute] -- [description]
        """        
        return self._get_element_details(parse_url(self._base_url, dispute_id), dispute_id, Dispute, response_type)

    def dispute_details_from_entity(self, dispute: Dispute, response_type: ResponseType=ResponseType.MINIMAL) -> PaypalApiResponse[Dispute]:
        """Performs an API call to show the details for a dispute, by ID. 
        
        Arguments:
            dispute_id {str} -- Dispute identifier
        
        Keyword Arguments:
            response_type {ResponseType} -- The preferred response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Dispute] -- [description]
        """        
        url = dispute.read_link
        headers = { 'Prefer': response_type.as_header_value() }
        api_response = self._execute_action_link(url, None, headers = headers)

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Dispute.serialize_from_json(api_response.json(), response_type))
    
    def _execute_basic_dispute_action(self, url: str, body: dict) -> PaypalApiBulkResponse[ActionLink]:
        """Executes a standard simple dispute action
        
        Arguments:
            url {str} -- The action URL
            note {str} -- Notes & comments
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        response = self._session.post(url, json.dumps(body))

        if response.status_code != 200:
            return PaypalApiBulkResponse(True, response)
        
        return PaypalApiBulkResponse(False, response, ActionLink.serialize_from_json(response['links']))

    def _execute_evidence_multipart_request(self, url: str, json_part: dict, files: MIMEApplication) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to provide evidence on a multipart request
        
        Arguments:
            dispute_id {str} -- dispute identifier
            json_part {dict} -- Json part of the multipart request
            files {MIMEApplication} -- files to be appended
        
        Keyword Arguments:
            return_addr {PaypalPortableAddress} -- [description] (default: {None})
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- [description]
        """
        multipart = MIMEMultipart('related')        
        text = 'input={};type=application/json'.format(json.dumps(json_part))

        json_part = MIMEText(text, 'json')
        # json_part = MIMEApplication(json_part,'json')

        multipart.attach(json_part)

        for f in files:
            multipart.attach(f)

        body = multipart.as_string()
        headers = dict(multipart.items())
        response = self._session.post(url, body, headers = headers)

        if response.status_code != 200:
            return PaypalApiBulkResponse(True, response)
        
        return PaypalApiBulkResponse(False, response, ActionLink.serialize_from_json(response['links']))

    def accept_claim(self, dispute_id: str, refund_amount: Money= None, **kwargs) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the paypal API to accept a claim & close the dispute in favor of the customer
        
        Arguments:
            dispute_id {str} -- [description]

        Keyword Arguments:
            note {str} -- Notes & comments
            accept_claim_reason {str} -- Reason to accept the claim (default REASON_NOT_SET)
            invoice_id {str} -- The merchant-provided ID of the invoice for the refund. 
            return_shipping_address {PaypalPortableAddress} --  The return address for the item.
            refund_amount {Money} -- To accept a customer's claim, the amount that the merchant agrees to refund the customer.
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        if refund_amount:
            kwargs['refund_amount'] = refund_amount.to_dict()
            self._remove_null_entries(kwargs['refund_amount'])
        
        if 'return_shipping_address' in kwargs.keys() and isinstance(kwargs['return_shipping_address'], PaypalPortableAddress):
            kwargs['return_shipping_address'] = kwargs['return_shipping_address'].to_dict()
            self._remove_null_entries(kwargs['return_shipping_address'])

        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id, 'accept-claim'), kwargs)

    def accept_offer(self, dispute_id: str, note: str) -> PaypalApiBulkResponse[ActionLink]:
        """Performs an API call to execute a customer acceptance of an offer from the merchant to resolve a dispute.
        
        Arguments:
            dispute_id {str} -- The dispute identifier 
            note {str} -- The customer notes about accepting of offer.
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id, 'accept-offer'), {'note': note})

    def acknowledge_returned_item(self, dispute_id: str, note: str) -> PaypalApiBulkResponse[ActionLink]:
        """Performs an API call to acknowledge that the customer returned an item for a dispute.
        
        Arguments:
            dispute_id {str} -- The dispute identifier 
            note {str} -- Merchant provided notes.
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id, 'acknowledge-return-item'), {'note': note})

    def appeal_dispute(self, dispute_id: str, evidence: DisputeEvidence, files: MIMEApplication, return_addr: PaypalPortableAddress=None) -> PaypalApiBulkResponse[ActionLink]:
        """Appeals a dispute
        
        Arguments:
            dispute_id {str} -- Dispute identifier
            evidence {DisputeEvidence} -- Appeal evidence 
            files {List[bytes]} -- files to be uploaded (current limit 10MB, 5 max per file)
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- [description]
        """
        url = parse_url(self._base_url, dispute_id, 'appeal')

        json_part = dict()
        json_part['evidences'] = evidence.to_dict()
        
        if return_addr: 
            json_part['return_address'] = return_addr.to_dict()
            self._remove_null_entries(json_part['return_address'])

        return self._execute_evidence_multipart_request(url, json_part, files)

    def deny_offer(self, dispute_id: str, note: str) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to deny an offer that the merchant proposes for a dispute.
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            note {str} -- Customer notes about the denial of offer
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id,'deny-offer'), {'note':note})

    def escalate_to_claim(self, dispute_id: str, note: str) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to escalate the dispute, by ID, to a PayPal claim.
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            note {str} -- Customer notes about the escalation
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id,'deny-offer'), {'note':note})

    def make_return_offer(self, dispute_id: str, note: str, offer_type: str, invoice_id: str=None, offer_amt: Money=None, return_addr: PaypalPortableAddress=None) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to make an offer to the other party to resolve a dispute
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            note {str} -- Customer notes about the escalation
            offer_type {str} -- The merchant-proposed offer type for the dispute

        Keyword Arguments:
            offer_amt {Money} --  The amount proposed to resolve the dispute. 
            invoice_id {str} -- The merchant-provided ID of the invoice for the refund. 
            return_shipping_address {PaypalPortableAddress} --  The return address for the item.
            refund_amount {Money} -- To accept a customer's claim, the amount that the merchant agrees to refund the customer.

        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        body = { 'note': note, 'offer_type' : offer_type }

        if invoice_id:
            body['invoice_id'] = invoice_id
        
        if offer_amt:
            body['offer_amount'] = offer_amt.to_dict()
            self._remove_null_entries(body['offer_amount'])
        
        if return_addr:
            body['return_address'] = return_addr.to_dict()
            self._remove_null_entries(body['return_address'])

        return self._execute_basic_dispute_action(parse_url(self._base_url, dispute_id,'make-offer'), body)

    def provide_evidence(self, dispute_id: str, evidence: DisputeEvidence, files: MIMEApplication, return_addr: PaypalPortableAddress=None) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to provide evidence 
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            evidence {DisputeEvidence} -- The evidence to be supported
            files {MIMEApplication} -- Files regarding the evidence as MIMEApplication
        
        Keyword Arguments:
            return_addr {PaypalPortableAddress} -- Portable return addr if needed (default: {None})
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        url = parse_url(self._base_url, dispute_id, 'provide-evidence')

        json_part = dict()
        json_part['evidences'] = evidence.to_dict()
        
        if return_addr: 
            json_part['return_address'] = return_addr.to_dict()
            self._remove_null_entries(json_part['return_address'])

        return self._execute_evidence_multipart_request(url, json_part, files)

    def provide_supporting_info(self, dispute_id: str, notes: str, supporting_doc: MIMEApplication) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the API to provide supporting information on a dispute
        
        Arguments:
            dispute_id {str} -- The dispute identifier
            notes {str} -- Merchant notes
            supporting_doc {MIMEApplication} -- Supporting document
                
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        url = parse_url(self._base_url, dispute_id, 'provide-supporting-info')
        return self._execute_evidence_multipart_request(url, {'notes': notes}, supporting_doc)

    def send_message_to_third_party(self, dispute_id: str, message: str) -> PaypalApiBulkResponse[ActionLink]:
        """Calls the api to send a message about a dispute, by ID, to the other party in the dispute
        
        Arguments:
            dispute_id {str} -- The dispute id
            message {str} -- Message to be sent
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- action links related to the dispute
        """
        url = parse_url(self._base_url, dispute_id, 'send-message')
        return self._execute_basic_dispute_action(url, {'message': message})

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)


"""
    Module with invoice templete related entities
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import ( 
    T,
    Item,   
    Money,
    ActionLink,
    PayPalEntity,
    ResponseType,
    RecipientInfo
)

from pypaypal.entities.invoicing.invoice import (
    MetaData,
    InvoicerInfo,
    FileReference,
    PartialPayment
)

class InvoiceListRequestField(Enum):
    ALL = 1
    NONE = 2

class TemplateDetail(PayPalEntity):
    """Template detail object representation.
    """

    def __init__(
        self, reference: str, currency_code: str, note: str, 
        terms_and_conditions: str, memo: str, payment_term: str,
        metadata: MetaData, attachments: List[FileReference] = [],
        **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.reference = reference
        self.currency_code = currency_code
        self.note  = note
        self.terms_and_conditions = terms_and_conditions
        self.memo = memo
        self.payment_term = payment_term
        self.metadata = metadata
        self.attachments = attachments or []

    def to_dict(self) -> dict:
        ret = super().to_dict()        
        if self.metadata:
            ret['metadata'] = [ x.to_dict() for x in self.metadata ]
        if self.attachments:
            ret['attachments'] = [ x.to_dict() for x in self.attachments ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        metadata, attachments = None, []
        
        if 'metadata' in json_data.keys():
            metadata = MetaData.serialize_from_json(json_data, response_type)
        
        if 'attachments' in json_data.keys():
            attachments = [FileReference.serialize_from_json(x, response_type) for x in json_data['attachments']]

        return cls(
            json_data.get('reference'), json_data.get('currency_code'), json_data.get('note'), 
            json_data.get('terms_and_conditions'), json_data.get('memo'), json_data.get('payment_term'),
            metadata, attachments, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(
        cls, currency_code: str, *, reference: str, note: str, 
        terms_and_conditions: str, memo: str, payment_term: str, 
        metadata: MetaData, attachments: List[FileReference] = []
    ):
        return cls(
            reference, currency_code, note, terms_and_conditions, 
            memo, payment_term, metadata, attachments or []
        )

class TemplateConfiguration(PayPalEntity):
    """Invoice configuration object representation
    """

    def __init__(self, tax_calc_after_discount: bool, tax_inclusive: bool, allow_tip: bool, partial_payment: PartialPayment=None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.allow_tip = allow_tip
        self.tax_inclusive = tax_inclusive
        self.partial_payment = partial_payment
        self.tax_calc_after_discount = tax_calc_after_discount
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        partial_payment = PartialPayment.serialize_from_json(json_data['partial_payment']) if 'partial_payment' in json_data.keys() else None

        return cls(
            json_data.get('tax_calc_after_discount'), json_data.get('tax_inclusive'),
            json_data.get('allow_tip'), partial_payment, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(
        cls, *, tax_calc_after_discount: bool = True, tax_inclusive: bool = False,
        allow_tip: bool = False, partial_payment: PartialPayment=None
    ):
        return cls(tax_calc_after_discount, tax_inclusive, allow_tip, partial_payment)

class TemplateInfo(PayPalEntity):
    """ Template info object representation. 
        Includes invoicer business information, invoice recipients,
        items, and configuration.
    """
    
    def __init__(
        self, detail: TemplateDetail, invoicer: InvoicerInfo, primary_recipients: RecipientInfo,
        items: List[Item], configuration: TemplateConfiguration, amount: Money, due_amount: Money,
        **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.detail = detail
        self.invoicer = invoicer
        self.primary_recipients = primary_recipients
        self.items = items
        self.configuration = configuration
        self.amount = amount
        self.due_amount = due_amount
        self.additional_recipients = self._json_response.get('additional_recipients', kwargs.get('additional_recipients'))
    
    def to_dict(self) -> dict:
        ret = super().to_dict()        
        if self.items:
            ret['items'] = [ x.to_dict() for x in self.items ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        primary_recipients, items  = [], []
        detail, invoicer, configuration, amount, due_amount  = None, None, None, None, None

        if 'primary_recipients' in json_data.keys():
            primary_recipients = [RecipientInfo.serialize_from_json(x, response_type) for x in json_data['primary_recipients']]
        if 'items' in json_data.keys():
            items = [Item.serialize_from_json(x, response_type) for x in json_data['items']]
        
        if 'detail' in json_data.keys():
            detail = TemplateDetail.serialize_from_json(json_data['detail'], response_type)
        if 'invoicer' in json_data.keys():
            invoicer = InvoicerInfo.serialize_from_json(json_data['invoicer'], response_type)
        if 'configuration' in json_data.keys():
            configuration = TemplateConfiguration.serialize_from_json(json_data['configuration'], response_type)
        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'], response_type)
        if 'due_amount' in json_data.keys():
            due_amount = Money.serialize_from_json(json_data['due_amount'], response_type)

        return cls(
            detail, invoicer, primary_recipients, items, configuration, amount, 
            due_amount, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(
        cls: Type[T], *, detail: TemplateDetail, invoicer: InvoicerInfo, primary_recipients: List[RecipientInfo], 
        additional_recipients: List[str], items: List[Item], configuration: TemplateConfiguration, amount: Money,
        due_amount: Money = None)  -> 'TemplateInfo':

        add_rep = [{ 'email_address': x} for x in additional_recipients]

        return cls(
            detail, invoicer, primary_recipients, items, 
            configuration, amount, due_amount, additional_recipients = add_rep
        )

class DisplayPreference(PayPalEntity):
    """Display preference object representation
    """
    def __init__(self, hidden: bool, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.hidden = hidden

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data['hidden'], json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, hidden: bool) -> 'DisplayPreference':
        return cls(hidden)

class TemplateItemSetting(PayPalEntity):
    """Template item setting object representation
    """
    def __init__(self, field_name: str, display_preference: DisplayPreference, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.field_name = field_name
        self.display_preference = display_preference

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        pref = None
        
        if 'display_preference' in json_data.keys():
            pref = DisplayPreference.serialize_from_json(json_data['display_preference'], response_type)

        return cls(json_data['field_name'], pref, json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, field_name: str, hidden: bool = False) -> 'TemplateItemSetting':
        return cls(field_name, DisplayPreference.create(hidden))

class TemplateSubtotalSetting(PayPalEntity):
    """Template subtotal setting object representation
    """
    def __init__(self, field_name: str, display_preference: DisplayPreference, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.field_name = field_name
        self.display_preference = display_preference

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        pref = None
        
        if 'display_preference' in json_data.keys():
            pref = DisplayPreference.serialize_from_json(json_data['display_preference'], response_type)

        return cls(json_data['field_name'], pref, json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, field_name: str, hidden: bool = False) -> 'TemplateSubtotalSetting':
        return cls(field_name, DisplayPreference.create(hidden))

class TemplateSettings(PayPalEntity):
    """Template show/hide settings object representation.
    """
    def __init__(self, 
        template_item_settings: List[TemplateItemSetting], 
        template_subtotal_settings: List[TemplateSubtotalSetting], **kwargs
        ):
            super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
            self.template_item_settings = template_item_settings
            self.template_subtotal_settings = template_subtotal_settings
    
    def to_dict(self) -> dict:
        ret = super().to_dict()        
        if self.template_item_settings:
            ret['template_item_settings'] = [ x.to_dict() for x in self.template_item_settings ]
        if self.template_subtotal_settings:
            ret['template_subtotal_settings'] = [ x.to_dict() for x in self.template_subtotal_settings ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        item_settings, subtotal_settings = [], []
        
        if 'template_item_settings' in json_data.keys():
            item_settings = [TemplateItemSetting.serialize_from_json(x, response_type) for x in json_data['template_item_settings']]

        if 'template_subtotal_settings' in json_data.keys():
            item_settings = [TemplateSubtotalSetting.serialize_from_json(x, response_type) for x in json_data['template_subtotal_settings']]

        return cls(item_settings, subtotal_settings, json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, field_name: str, hidden: bool = False) -> 'TemplateSubtotalSetting':
        return cls(field_name, DisplayPreference.create(hidden))

class Template(PayPalEntity):
    """Invoice template object representation    
    """
    def __init__(
        self, name: str, default_template: bool, 
        template_info: TemplateInfo, settings: TemplateSettings, 
        unit_of_measure: str, **kwargs
        ):
            super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
            self.name = name
            self.settings = settings
            self.template_info = template_info
            self.unit_of_measure = unit_of_measure
            self.default_template = default_template
            self.id = self._json_response.get('id', kwargs.get('id'))
            self.standard_template = self._json_response.get('standard_template', kwargs.get('standard_template'))
            self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def replace_link(self) -> ActionLink:
        """Retrieves a link to replace (update) this template.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'replace', self.links), None)

    @property
    def delete_link(self) -> ActionLink:
        """Retrieves a link to delete this template.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'delete', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        template_info, template_settings = None, None
        
        if 'template_info' in json_data.keys():
            template_info = TemplateInfo.serialize_from_json(json_data['template_info'], response_type)
        if 'template_settings' in json_data.keys():
            template_settings = TemplateSettings.serialize_from_json(json_data['template_settings'], response_type)

        return cls(
            json_data.get('name'), json_data.get('default_template'), template_info, 
            template_settings, json_data.get('unit_of_measure'), json_response= json_data,
            response_type = response_type
        )

    @classmethod
    def create(
            cls, name: str, template_info: TemplateInfo, settings: TemplateSettings, 
            unit_of_measure: str, default_template: bool = False
        ) -> 'Template':
            return cls(name, default_template, template_info, settings, unit_of_measure)



"""
    Payment refunds resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/payments/v2/    
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LIVE_API_BASE_URL,
    SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    ResponseType,
    PaypalApiResponse,
)

from pypaypal.entities.payments.refunds import Refund

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LIVE_API_BASE_URL, 'payments', 'refunds')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(SANDBOX_API_BASE_URL, 'payments', 'refunds')

T = TypeVar('T', bound = 'RefundClient')

class RefundClient(ClientBase):
    """Refunds resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def show_refund_details(self, refund_id: str) -> PaypalApiResponse[Refund]:
        """Calls the paypal API to show details for an authorized payment, by ID.
        
        Arguments:
            refund_id {str} -- The refund id
        
        Returns:
            PaypalApiResponse[Refund] -- An api response with the refund details
        """
        api_response = self._session.get(parse_url(self._base_url, refund_id))

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Refund.serialize_from_json(api_response.json()))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)



"""
    Module with all referenced payout related entities.
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import ( 
    T, 
    Money,
    ActionLink, 
    PaypalName,
    PayPalEntity, 
    ResponseType, 
    PaypalPhoneDetail
)

class ProcessingStatus(Enum):
    # The item is pending.
    PENDING = 1
    # The item is processing.
    PROCESSING = 2
    # The item succeeded.
    SUCCESS = 3
    # The item failed.
    FAILED = 4
    # The payout failed for the item.
    PAYOUT_FAILED = 5

class ProcessingStatusReason(Enum):
    # An internal error occurred.
    INTERNAL_ERROR = 1
    # The balance is not enough.
    NOT_ENOUGH_BALANCE = 2
    # The amount check failed.
    AMOUNT_CHECK_FAILED = 3
    # A merchant-partner permissions issue occurred.
    MERCHANT_PARTNER_PERMISSIONS_ISSUE = 4
    # The merchant has restrictions.
    MERCHANT_RESTRICTIONS = 5
    # The transaction is under dispute.
    TRANSACTION_UNDER_DISPUTE = 6
    # The transaction is not valie.
    TRANSACTION_NOT_VALID = 7
    # The currency is not supported.
    UNSUPPORTED_CURRENCY = 8
    # The payout was already initiated.
    PAYOUT_INITIATED = 9
    # The payout was already completed.
    PAYOUT_ALREADY_COMPLETED_FOR_REFERENCE = 10

class ReferenceType(Enum):
    # The reference type is a transaction ID.
    TRANSACTION_ID = 1
    # The reference type is other.
    OTHERS = 2

class ExecutionType(Enum):
    SYNC = 1
    ASYNC = 2

    @property
    def is_sync(self) -> bool:
        return self.value == 1

    @property
    def is_async(self) -> bool:
        return self.value == 2

class ProcessingState(PayPalEntity):
    """Referenced payout item processing state obj definition
    """
    def __init__(self, status: str = None, reason: str = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.status = status
        self.reason = reason
    
    @property
    def status_enum(self) -> ProcessingStatus:
        try:
            return ProcessingStatus[self.status] if self.status else None
        except:
            return None
    
    @property
    def reason_enum(self) -> ProcessingStatusReason:
        try:
            return ProcessingStatusReason[self.reason] if self.reason else None
        except:
            return None

    @classmethod
    def create(cls, *, status: ProcessingStatus = None, reason: str = ProcessingStatusReason) -> 'ProcessingState':
        return cls(
            status = status.name if status else None, 
            reason = reason.name if reason else None
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class PayoutDirective(PayPalEntity):
    """Payout Directive obj representation
    """
    def __init__(self, financial_instrument_id: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.financial_instrument_id = financial_instrument_id
    
    @classmethod
    def create(cls, financial_instrument_id: str) -> 'PayoutDirective':
        return cls(financial_instrument_id = financial_instrument_id)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(**json_data, json_response = json_data, response_type = response_type)

class ReferencedPayoutsItem(PayPalEntity):
    """Referenced Payouts Item obj representation
    """

    # Special entity types
    _ENTITY_TYPES = { 'processing_state': ProcessingState, 'payout_amount': Money }

    def __init__(
            self, item_id: str = None, processing_state: str = None, 
            reference_id: str = None, reference_type: str = None, payout_transaction_id: str = None, 
            disbursement_transaction_id: str = None, external_merchant_id: str = None, 
            payee_email: str = None, payout_amount: Money = None, payout_destination: str = None, 
            invoice_id: str = None, custom: str = None, **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.custom = custom
        self.item_id = item_id
        self.invoice_id = invoice_id
        self.payee_email = payee_email
        self.reference_id = reference_id
        self.payout_amount = payout_amount
        self.reference_type = reference_type
        self.processing_state = processing_state
        self.payout_destination = payout_destination
        self.external_merchant_id = external_merchant_id
        self.payout_transaction_id = payout_transaction_id
        self.disbursement_transaction_id = disbursement_transaction_id
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def reference_type_enum(self) -> ReferenceType:
        try:
            return ReferenceType[self.reference_type] if self.reference_type else None
        except:
            return None

    @classmethod
    def create(
        cls, *, item_id: str = None, processing_state: ProcessingState = None, 
        reference_id: str = None, reference_type: ReferenceType = None,
        disbursement_transaction_id: str = None, external_merchant_id: str = None, 
        payee_email: str = None, payout_amount: Money = None, 
        invoice_id: str = None, custom: str = None
    ) -> 'ReferencedPayoutsItem': 
        return cls(
            reference_type = reference_type.name if reference_type else None, payout_amount = payout_amount, 
            processing_state = processing_state, reference_id = reference_id, payee_email = payee_email,
            disbursement_transaction_id = disbursement_transaction_id, item_id = item_id, custom = custom,
            external_merchant_id = external_merchant_id, invoice_id = invoice_id,
        )

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        args = {
            **json_data,            
            **{ k : cls._ENTITY_TYPES[k].serialize_from_json(v, response_type) for k,v in json_data.items() if k in cls._ENTITY_TYPES.keys() and v }
        }

        return cls(**args, json_response = json_data, response_type = response_type)

class ReferencedPayoutCall(PayPalEntity):
    """ Class to wrap reference payout requests / responses
    """
    def __init__(
            self, payout_directive: PayoutDirective = None, referenced_payouts: List[ReferencedPayoutsItem] = [], **kwargs
        ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.payout_directive = payout_directive
        self.referenced_payouts = referenced_payouts or []
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return None

class ReferencedPayoutRequest(ReferencedPayoutCall):
    """Class to wrap reference payout requests    
    """
    def __init__(
            self, payout_directive: PayoutDirective, referenced_payouts: List[ReferencedPayoutsItem], 
            execution_type: ExecutionType = ExecutionType.SYNC 
        ):
        super().__init__(payout_directive = payout_directive, referenced_payouts = referenced_payouts, execution_type = execution_type)
        self._execution_type = execution_type

    @property
    def execution_type(self) -> ExecutionType:
        return ExecutionType.SYNC if self._execution_type == ExecutionType.SYNC and len(self.referenced_payouts) <= 10 else ExecutionType.ASYNC
    
    @classmethod
    def create(
            cls, payout_directive: PayoutDirective, referenced_payouts: List[ReferencedPayoutsItem], 
            execution_type: ExecutionType = ExecutionType.SYNC
        ) -> 'ReferencedPayoutRequest':
        exec_type = ExecutionType.SYNC if execution_type == ExecutionType.SYNC and len(referenced_payouts) <= 10 else ExecutionType.ASYNC
        return cls(payout_directive, referenced_payouts, exec_type)

class ReferencedPayoutResponse(ReferencedPayoutCall):
    """Class to wrap reference payout requests    
    """
    def __init__(
            self, payout_directive: PayoutDirective,
            referenced_payouts: List[ReferencedPayoutsItem], **kwargs
        ):
        super().__init__(payout_directive = payout_directive, referenced_payouts = referenced_payouts, **kwargs)

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        payout_directive, referenced_payouts  = None,  []

        if 'payout_directive'  in json_data.keys():
            payout_directive = PayoutDirective.serialize_from_json(payout_directive, response_type)
        if 'referenced_payouts' in json_data.keys():
            referenced_payouts = [ReferencedPayoutsItem.serialize_from_json(x, response_type) for x in json_data['referenced_payouts']]
        
        return cls(
            payout_directive = payout_directive, referenced_payouts = referenced_payouts, 
            json_response = json_data, response_type = response_type
        )


"""
    Tracker resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/tracking/v1/ 
"""

import json
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase
from pypaypal.entities.base import ResponseType, PaypalApiResponse

from pypaypal.entities.trackers import Tracker

from pypaypal.http import ( 
    parse_url,
    PayPalSession,     
    LEGACY_LIVE_API_BASE_URL, 
    LEGACY_SANDBOX_API_BASE_URL 
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL,'shipping/trackers')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL,'shipping/trackers')

"""
    Supported tracker properties for tracking creation
"""
_TRACKER_PROPERTIES = { 
    'status', 'carrier', 'transaction_id', 'tracking_number', 'quantity', 'notify_buyer',
    'shipment_date', 'last_updated_time', 'carrier_name_other', 'postage_payment_id', 
    'tracking_number_validated' 
}

T = TypeVar('T', bound = 'TrackersClient')

class TrackersClient(ClientBase):
    """Trackers resource group client class
    """

    def __init__(self, url: str, session: PayPalSession):
        """Client ctor
        
        Arguments:
            session {PayPalSession} -- The http session 
        """
        super().__init__(url, session)

    def show_tracking_info(self, transaction_id: str, tracking_number: int, response_type: ResponseType= ResponseType.MINIMAL) -> PaypalApiResponse[Tracker]:
        """Gets the tracking info for a given tracker key
        
        Arguments:
            transaction_id {str} -- The id of the transaction
            tracking_number {int} -- The tracking number

        Keyword Arguments:
            response_type {ResponseType} -- [description] (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Tracker] -- Response container with a Tracker instance including the required information
        """     
        headers = { 'Prefer': response_type.as_header_value() } 
        url = parse_url(self._base_url, f'{transaction_id}-{tracking_number}')
        api_response = self._session.get(url, None, headers=headers)
        
        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response, Tracker.serialize_from_json(api_response.json(), response_type))

    def update_tracking_info(self, transaction_id: str, tracking_number: int, **kwargs) -> PaypalApiResponse[Tracker]:
        """Updates tracking info in the api
        
        Arguments:
            transaction_id {str} -- The id of the transaction
            tracking_number {int} -- The tracking number

        Keyword Arguments:
            mappings with the fields to be updated, see the docs for the names & values.
        
        Returns:
            PaypalApiResponse[Tracker] -- Response container with the response status
        """
        url = parse_url(self._base_url, f'{transaction_id}-{tracking_number}')
        api_response = self._session.put(url, json.dumps(kwargs))        
        error = api_response.status_code != 204

        return PaypalApiResponse(error, api_response)

    def add_trackers(self, trackers: List[Tracker], response_type: ResponseType= ResponseType.MINIMAL) -> PaypalApiResponse[Tracker]:
        """Adds trackers to a transaction
        
        Arguments:
            trackers {List[Tracker]} -- [description]
        
        Returns:
            PaypalApiResponse[Tracker] -- [description]
        """
        body = []
        url = parse_url(self._base_url, 'trackers-batch')
        headers = { 'Prefer': response_type.as_header_value() }

        for t in trackers:
            b = t.to_dict()
            b['shipment_date'] = b.pop('_shipment_date', None)
            b['last_updated_time'] = b.pop('_last_updated_time', None)
            self._clean_dictionary(b, _TRACKER_PROPERTIES)
            body.append(b)
        
        api_response = self._session.post(url, json.dumps(body), headers = headers)
        
        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response, Tracker.serialize_from_json(api_response.json(), response_type))

    def tracking_info_by_entity(self, tracker: Tracker, response_type: ResponseType= ResponseType.MINIMAL) -> PaypalApiResponse[Tracker]:
        """Gets the tracking info for a given tracking entity
        
        Arguments:
            tracker {Tracker} -- a tracking entity
        
        Returns:
            PaypalApiResponse[Tracker] -- Response container with a Tracker instance including the required information
        """
        url = tracker.read_link
        headers = { 'Prefer': response_type.as_header_value() }
        api_response = self._execute_action_link(url, None, headers = headers)

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response, Tracker.serialize_from_json(api_response.json(), response_type))

    def update_tracking_info_by_entity(self, tracker: Tracker) -> PaypalApiResponse[Tracker]:
        """Updates tracking info in the api
        
        Arguments:
            transaction_id {str} -- The id of the transaction
            tracking_number {int} -- The tracking number

        Keyword Arguments:
            mappings with the fields to be updated, see the docs for the names & values.
        
        Returns:
            PaypalApiResponse[Tracker] -- Response container with the response status
        """
        body = tracker.json_data
        url = tracker.update_link

        for item in tracker.to_dict().items():
            key = item.key
            if key in body.keys():
                body[key] = item.value

        api_response = self._execute_action_link(url, body)
        error = api_response.status_code != 204
        return PaypalApiResponse(error, api_response)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a tracker client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of TrackersClient with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)


"""
    Payment authorizations resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/payments/v2/ 
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LIVE_API_BASE_URL,
    SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    Money,
    ResponseType,
    PaypalApiResponse,
    PaymentInstruction,
    PaypalApiBulkResponse
)

from pypaypal.entities.payments.captures import Capture
from pypaypal.entities.payments.authorizations import Authorization

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LIVE_API_BASE_URL, 'payments', 'authorizations')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(SANDBOX_API_BASE_URL, 'payments', 'authorizations')


T = TypeVar('T', bound = 'AuthorizationClient')

class AuthorizationClient(ClientBase):
    """Authorizations resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)
    
    def show_authorization_details(self, authorization_id: str) -> PaypalApiResponse[Authorization]:
        """Calls the paypal API to show details for an authorized payment, by ID.
        
        Arguments:
            authorization_id {str} -- The authorization id
        
        Returns:
            PaypalApiResponse[Authorization] -- An api response with the authorization details
        """
        api_response = self._session.get(parse_url(self._base_url, authorization_id))

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Authorization.serialize_from_json(api_response.json()))
    
    def capture_authorized_payment(
        self, authorization_id: str, invoice_id: str = None, note_to_payer: str = None,
        instruction: PaymentInstruction = None, amount: Money = None, final_capture: bool = False,
        request_id: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Capture]:
        """Calls the paypal API to capture an authorized payment, by ID.
        
        Arguments:
            authorization_id {str} -- authorization id
            invoice_id {str} -- associated invoice
            note_to_payer {str} -- informational notes about the settlement. Appears in the payer's transaction history and received emails.
        
        Keyword Arguments:
            instruction {PaymentInstruction} -- Any additional payment instructions for PayPal for Partner customers (default: {None})
            amount {Money} -- The amount to capture. If none the full amount will be captured (default: {None}).
            final_capture {bool} -- dictates whether you can make additional captures against the authorized payment (default: {False}).
            request_id {str} -- Paypal request id for idempotence (default: {None})
            response_type {ResponseType} -- desired response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Authorization] -- An api response with the authorization details
        """
        body = dict()
        url = parse_url(self._base_url, authorization_id, 'capture')
        headers = {
            'Prefer': response_type.as_header_value()            
        }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        
        if invoice_id:
            body['invoice_id'] = invoice_id
        if note_to_payer:
            body['note_to_payer'] = note_to_payer
        if final_capture:
            body['final_capture'] = final_capture
        if instruction:
            body['instruction'] = instruction        
        if amount:
            body['amount'] = amount.to_dict()

        api_response = self._session.post(url, json.dumps(body), headers = headers)

        if api_response.status_code != 201:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Capture.serialize_from_json(api_response.json()))

    def reauthorize_payment(
        self, authorization_id: str, amount: Money, request_id: str = None,
        response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Authorization]:
        """Calls the paypal API to reauthorize an authorized payment, by ID. 
           See the api docs for a full set of rules.
        
        Arguments:
            authorization_id {str} -- The authorization id
            amount {Money} -- The amount to reauthorize for an authorized payment.
        Returns:
            PaypalApiResponse[Authorization] -- An api response with the authorization details
        """
        body = dict()
        url = parse_url(self._base_url, authorization_id, 'reauthorize')
        headers = {
            'Prefer': response_type.as_header_value()            
        }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        
        if amount:
            body['amount'] = amount.to_dict()

        api_response = self._session.post(url, json.dumps(body), headers = headers)

        if api_response.status_code != 201:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Authorization.serialize_from_json(api_response.json()))

    def void_payment(self, authorization_id: str, auth_assertion_token: str = None) -> PaypalApiResponse:
        """Calls the paypal API to void an authorized payment, by ID. 
           See the api docs for a full set of rules.
        
        Arguments:
            authorization_id {str} -- The authorization id

        Keyword Arguments:
            auth_assertion_token {str} -- PayPal-Auth-Assertion see: https://developer.paypal.com/docs/api/payments/v2/docs/api/reference/api-requests/#paypal-auth-assertion (default: {None})
            
        Returns:
            PaypalApiResponse -- An api response with the authorization details.
        """
        api_response = None
        url = parse_url(self._base_url, authorization_id, 'void')

        if auth_assertion_token:
            api_response = self._session.post(url, None, headers = {'PayPal-Auth-Assertion' : auth_assertion_token})
        else:
            api_response = self._session.post(url, None)

        if api_response.status_code != 204:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)



"""
    Paypal Web Expirience Profile resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/payment-experience/v1/#definition-presentation
"""

import json

from datetime import datetime
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL,
    LEGACY_SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.webexp import WebExpProfile, FlowConfig

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'payment-experience', 'web-profiles')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'payment-experience', 'web-profiles')

T = TypeVar('T', bound = 'WebExpClient')

class WebExpClient(ClientBase):
    """Paypal Web Expirience Profile resource group client class
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def create_profile(self, profile: WebExpProfile, request_id: str = None) -> PaypalApiResponse[WebExpProfile]:
        """Calls the paypal Api to create a WebExp Profile
        
        Arguments:
            profile {WebExpProfile} -- The profile information.
        
        Keyword Arguments:
            request_id {str} -- Paypal request id for idempotency (default: {None})
        
        Returns:
            PaypalApiResponse[WebExpProfile] -- An api response with the profile
        """        
        headers = {'PayPal-Request-Id': request_id} if request_id else dict()
        api_response = self._session.post(self._base_url, json.dumps(profile.to_dict()), headers = headers)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, WebExpProfile.serialize_from_json(api_response.json()))

    def list_profiles(self) -> PaypalApiBulkResponse[WebExpProfile]:
        """Calls the paypal Api to get a list with the latest 20 web experience profiles 
           for a merchant or subject.
        
        Returns:
            PaypalApiBulkResponse[WebExpProfile] -- Latest 20 profiles
        """
        api_response = self._session.get(self._base_url)

        if api_response.status_code != 200:
            return PaypalApiBulkResponse(True, api_response)

        return PaypalApiBulkResponse(False, api_response, [WebExpProfile.serialize_from_json(x) for x in api_response.json()])
    
    def delete_profile(self, profile_id: str) -> PaypalApiResponse:
        """Calls the paypal Api to delete a WebExp Profile
        
        Arguments:
            profile_id {WebExpProfile} -- The profile id.
        
        Returns:
            PaypalApiResponse -- An api response 
        """
        api_response = self._session.delete(parse_url(self._base_url, profile_id))

        if api_response.status_code != 204:
            return PaypalApiBulkResponse(True, api_response)
        return PaypalApiBulkResponse(False, api_response)
    
    def show_profile_details(self, profile_id: str) -> PaypalApiResponse[WebExpProfile]:
        """Calls the paypal Api to delete a WebExp Profile
        
        Arguments:
            profile_id {WebExpProfile} -- The profile id.
        
        Returns:
            PaypalApiResponse[WebExpProfile] -- An api response with the profile info
        """
        api_response = self._session.get(parse_url(self._base_url, profile_id))

        if api_response.status_code != 200:
            return PaypalApiBulkResponse(True, api_response)
        return PaypalApiBulkResponse(False, api_response, WebExpProfile.serialize_from_json(api_response.json()))

    def path_update_profile(self, profile_id: str, updates: List[PatchUpdateRequest]) ->  PaypalApiResponse:
        """Calls the api to partially-update a web experience profile, by ID.
        
        Arguments:
            profile_id {str} -- The profile id
            updates {List[PatchUpdateRequest]} -- list of update operations
        
        Returns:
            PaypalApiResponse -- API response status
        """
        url = parse_url(self._base_url, profile_id)
        body = json.dumps([{ 'op': x.operation, 'path': x.path, 'value': x.value } for x in updates ])

        api_response = self._session.patch(url, body)

        if api_response.status_code != 204:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response)

    def fully_update_profile(self, profile_id: str, name: str, temporary: bool, flow_config: FlowConfig) -> PaypalApiResponse:
        """Calls the api to fully update a web experience profile, by ID.
        
        Arguments:
            profile_id {str} -- The profile id
            updates {List[PatchUpdateRequest]} -- list of update operations
        
        Returns:
            PaypalApiResponse -- API response status
        """
        url = parse_url(self._base_url, profile_id)
        body = WebExpProfile.create(name, temporary, flow_config)

        api_response = self._session.put(url, body)

        if api_response.status_code != 204:
            return PaypalApiResponse(True, api_response)
        return PaypalApiResponse(False, api_response)



"""
    Plan resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/subscriptions/v1/
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List, NamedTuple

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL,
    LEGACY_SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    PaypalPage,
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.subscriptions.plans import Plan, PricingScheme

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'billing', 'plans')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'billing', 'plans')

T = TypeVar('T', bound = 'PlanClient')

class UpdatePricingSchemeRequest(NamedTuple):
    """Immutable object to send pricing scheme update requests to the API.
    """
    billing_cycle_sequence: int
    pricing_scheme: PricingScheme

    def to_dict(self) -> dict:
        return { 
            'pricing_scheme': self.pricing_scheme,
            'billing_cycle_sequence': self.billing_cycle_sequence            
        }

class PlanClient(ClientBase):
    """Plans resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def create_plan(self, plan: Plan, request_id: str = None, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Plan]:
        """Calls the paypal API to create a plan
        
        Arguments:
            plan {Plan} -- The plan to be made (use the create factory method)
        
        Keyword Arguments:
            request_id {str} -- Paypal request id for idempotence (default: {None})
            response_type {ResponseType} -- Peferred response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Plan] -- an API response object with the plan info
        """
        url = self._base_url
        body = json.dumps(plan.to_dict())
        headers = { 'Prefer': response_type.as_header_value() }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        
        api_response = self._session.post(url, body, headers = headers)

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, Plan.serialize_from_json(api_response.json(), response_type))

    def list_plans(
            self, product_id: str = None, plan_ids: List[str] = [], 
            page_size: int = 10, page: int = 1, total_required: bool = True, 
            response_type: ResponseType = ResponseType.MINIMAL
        ) -> PaypalPage[Plan]:
        """Calls the Paypal API to list different plan details in a page
        
        Keyword Arguments:
            product_id {str} -- product id to query (default: {None})
            plan_ids {List[str]} -- list of desired plan ids (10 supported) (default: {[]})
            page_size {int} -- size of the page (default: {10})
            page {int} -- desired page (default: {1})
            total_required {bool} -- flag to show the total count in the response (default: {True})
            response_type {ResponseType} -- response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalPage[Plan] -- Page with plan details
        """
        url = self._base_url
        params = { 'page_size': page_size, 'page': page, 'total_required': total_required }

        if product_id:
            params['product_id'] = product_id
        if plan_ids:
            params['plan_ids'] = ','.join(plan_ids)

        api_response = self._session.get(url, params, headers = { 'Prefer': response_type.as_header_value() })

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalPage.full_parse_success(api_response, Plan, 'plans', response_type)

    def list_plans_from_link(self, link: ActionLink, response_type: ResponseType = ResponseType.MINIMAL) -> PaypalPage[Plan]:
        """Calls the Paypal API to list different plan details in a page

        Arguments:
            link {ActionLink} -- Page link

        Keyword Arguments:
            response_type {ResponseType} -- response type (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalPage[Plan] -- Page with plan details
        """
        url = link.href
        api_response = self._session.get(url, headers = { 'Prefer': response_type.as_header_value() })

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalPage.full_parse_success(api_response, Plan, 'plans', response_type)

    def update_plan(self, plan_id: str, update_request: List[PatchUpdateRequest]) -> PaypalApiResponse:
        """Patch request to update a plan. See the docs for details 
        
        Arguments:
            plan_id {str} -- id of the plan to be updated.
            update_request {PatchUpdateRequest} -- request object with update info.
        
        Returns:
            PaypalApiResponse -- Response obj with operation status
        """
        url = parse_url(self._base_url, plan_id) 
        body = json.dumps([ {'op': x.operation, 'value': x.value, 'path': x.path } for x in update_request ])

        api_response = self._session.patch(url, body)

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    def show_plan_details(self, plan_id: str) -> PaypalApiResponse[Plan]:
        """Calls the paypal API to get the plan details
        
        Arguments:
            plan_id {str} -- The plan identifier
        
        Returns:
            PaypalApiResponse[Plan] -- Api response obj with the plan info.
        """
        api_response = self._session.get(parse_url(self._base_url, plan_id))

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalApiResponse.success(api_response, Plan.serialize_from_json(api_response.json()))

    def activate_plan(self, plan_id: str) -> PaypalApiResponse:
        """Calls the API to activate a plan
        
        Arguments:
            plan_id {str} -- The id of the plan to activate
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        api_response = self._session.post(parse_url(self._base_url, plan_id, 'activate'))

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    def deactivate_plan(self, plan_id: str) -> PaypalApiResponse:
        """Calls the API to deactivate a plan
        
        Arguments:
            plan_id {str} -- The id of the plan to deactivate
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        api_response = self._session.post(parse_url(self._base_url, plan_id, 'deactivate'))

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    def update_pricing(self, plan_id: str, pricing_schemes: List[UpdatePricingSchemeRequest]) -> PaypalApiResponse:
        """Calls the API to update a plan's pricing scheme
        
        Arguments:
            plan_id {str} -- The id of the plan to deactivate
            pricing_schemes List[UpdatePricingSchemeRequest] -- The pricing scheme updates
        
        Returns:
            PaypalApiResponse -- Response with the operation status
        """
        body = json.dumps([ x.to_dict() for x in pricing_schemes ])
        url = parse_url(self._base_url, plan_id, 'update-pricing-schemes')
        
        api_response = self._session.post(url, body)

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalApiResponse.success(api_response)

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)



"""
    Base module with common needs for every resource client
"""

from typing import Type

from pypaypal.http import PayPalSession

from pypaypal.entities.base import ( 
    T, 
    ActionLink, 
    PaypalPage, 
    ResponseType,
    RequestMethod, 
    PaypalApiResponse
)

class ClientBase:
    """
        Base client class for every resource client.
    """
    def __init__(self, base_url: str, session: PayPalSession):
        """Class ctor
        
        Arguments:
            base_url {str} -- The base url for the resource group
            session {PayPalSession} -- The paypal session that will perform the requests
        """
        self._session = session
        self._base_url = base_url
    
    def _get_element_details(self, url: str, element_id: str, element_class: Type[T], response_type: ResponseType) -> PaypalApiResponse[T]:
        """Calls the API to get an element detail by it's id
        
        Arguments:
            element_id {str} -- The element identifier
            element_class {[type]} -- serializable class reference
        
        Returns:
            PaypalApiResponse[T] -- A response wrapper with the object & operation result
        """

    def _page_from_link(self, link: ActionLink, element_key: str, element_class: Type[T]) -> PaypalPage[T]:
        """Performs an API call to get a page of elements
        
        Arguments:
            link {ActionLink} -- Page action link (HATEOAS)
            element_key {str} -- The key for the element lists inside the page json
            element_class {T} -- class reference of a PayPalEntity subclass
        Returns:
            PaypalPage[T] -- A page with the desired elements
        """
        api_response = self._execute_action_link(link, None)

        if api_response.status_code != 200:
            return PaypalPage(True, api_response, 0, 0, [], [])
        
        json_response = api_response.json()
        total_items = json_response.get('total_items', 0)
        total_pages = json_response.get('total_pages', 0)
        elements = [ element_class.serialize_from_json(x) for x in json_response[element_key]]
        links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in json_response['links']]

        return PaypalPage(False, api_response, total_items, total_pages, elements, links)

    def _remove_null_entries(self, dictionary: dict):
        """Cleans a dictionary removing null entries and invalid keys
        
        Arguments:
            dictionary {dict} -- The dictionary to be checked
            valid_keys {set} -- The keys that should not be deleted
        """
        keys = list(dictionary.keys())
        for k in keys:
            if dictionary.get(k) == None:
                del dictionary[k]

    def _clean_dictionary(self, dictionary: dict, valid_keys: set):
        """Cleans a dictionary removing null entries and invalid keys
        
        Arguments:
            dictionary {dict} -- The dictionary to be checked
            valid_keys {set} -- The keys that should not be deleted
        """
        keys = list(dictionary.keys())
        for k in keys:
            if dictionary.get(k) == None or not k in valid_keys:
                del dictionary[k]

    def _execute_action_link(self, link: ActionLink, body: str, **kwargs):
        """Executes an action link from a given entity
        
        Supported kwargs: 
            params for query string parameters in GET requests
            every other kwarg supported by the requests library

        Arguments:
            link {ActionLink} -- The action link to be executed
            body {str} -- the action link body if needed
        """
        url = link.href
        method = link.method

        if method == RequestMethod.POST:
            return self._session.post(url, body, **kwargs)
        elif method == RequestMethod.PUT:
            return self._session.put(url, body, **kwargs)
        elif method == RequestMethod.PATCH:
            return self._session.patch(url, body, **kwargs)
        elif method == RequestMethod.DELETE:
            return self._session.delete(url, **kwargs)
        else: 
            # Defaulting to get
            return self._session.get(url, kwargs.pop('params',None), **kwargs)